/* ===================================================================== */
/*
 * This file is part of CARDAMOM (R) which is jointly developed by THALES
 * and SELEX-SI. It is derivative work based on PERCO Copyright (C) THALES
 * 2000-2003. All rights reserved.
 * Copyright (C) THALES 2004-2006. All rights reserved.
 * 
 * CARDAMOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * CARDAMOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with CARDAMOM; see the file COPYING. If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/* ===================================================================== */


#include <Foundation/common/Assert.hpp>
#include <CCMContainer/idllib/CdmwCcmContainer.stub.hpp>
#include <CCMContainer/ccmcif/Context.hpp>
#include <Foundation/orbsupport/CORBAMessaging.hpp>
#include <CCMContainer/ccmcif/SessionContext.hpp>
#include <CCMContainer/ccmcommon/CCMUtils.hpp>
#include <Foundation/orbsupport/ExceptionMinorCodes.hpp>
#include <CCMContainer/idllib/CdmwDeployment.stub.hpp>
#include <CCMContainer/idllib/CdmwCcmCif.stub.hpp>

namespace Cdmw {

namespace CCM {

namespace CIF {

/*
 * This class is the base class for ccmcif context embedding specific CCM 
 * context.
 */
Context::Context(CdmwCcmContainer::CCM2Context_ptr ctx)
  : m_ccm_context(CdmwCcmContainer::CCM2Context::_duplicate(ctx))
{
  CDMW_ASSERT(!CORBA::is_nil(ctx));
}

Context::~Context() 
{
  m_ccm_context = CdmwCcmContainer::CCM2Context::_nil();
}

CORBA::Object_ptr 
Context::resolve_initial_references(const char* identifier)
  throw (CdmwCcmContainer::CCM2Context::InvalidName,
         CORBA::SystemException)
{
  return m_ccm_context->resolve_initial_references(identifier);
}

Components::CCMHome_ptr 
Context::get_CCM_home() throw(CORBA::SystemException)
{
  return m_ccm_context->get_CCM_home();
}
   
::CosEventChannelAdmin::EventChannel_ptr
Context::obtain_event_channel_for_emitter(::Components::PortDescription* port)
  throw (CORBA::SystemException)
{
  return m_ccm_context->obtain_event_channel_for_emitter(port);
}

::CosEventChannelAdmin::EventChannel_ptr
Context::create_event_channel_for_publisher(::Components::PortDescription* port)
  throw (CORBA::SystemException)
{
  return m_ccm_context->create_event_channel_for_publisher(port);
}

void Context::set_config_values( const std::string& comp_oid,
                                 const Components::ConfigValues& config )
  throw( Cdmw::Common::AlreadyExistException )
{

  ConfigValuesMap::iterator cv_itr = m_config_values.find( comp_oid );
  
  if( cv_itr != m_config_values.end() )
    throw Cdmw::Common::AlreadyExistException();

  // create a copy of config values to be stored in the map
  m_config_values.insert( ConfigValuesMap::value_type( comp_oid,  new Components::ConfigValues(config)));

  PRINT_INFO("insertion of config values in the map of the context. "
             << "Key member of the map : oid = "
             << comp_oid);
}

Components::ConfigValues* Context::get_config_values( const std::string& comp_oid )
  throw( Cdmw::Common::NotFoundException )
{
  ConfigValuesMap::iterator cv_itr = m_config_values.find( comp_oid );
  Components::ConfigValues_var config_values;

  if( cv_itr != m_config_values.end() ){
    config_values = cv_itr->second;
    PRINT_INFO("Context get_config_values : config values found.");
    return config_values._retn();
  }else{
    PRINT_INFO("Context get_config_values : config values not found.");
    throw Cdmw::Common::NotFoundException();
  }
}

CORBA::Object_ptr 
Context::set_timeout_policy ( unsigned long timeout_in_ms,
                              CORBA::Object_ptr obj )
  throw ( CORBA::SystemException, CORBA::PolicyError )
{

  /* WARNING : TimeT is expressed in units of 100 nano seconds */
  TimeBase::TimeT _timeout = ((unsigned long) (timeout_in_ms * 10000));

  PRINT_INFO("Context set_timeout_policy: timeout = " << _timeout << " ( * 100 ns)");

  CORBA::Any timeout_any;
  timeout_any <<= _timeout;
    
  CORBA::PolicyList timeout_policy_list (1);
  timeout_policy_list.length (1);
    
  /*
   * recover the orb reference
   */
  CdmwCcmContainer::Session2Context_var session2context = CdmwCcmContainer::Session2Context::_narrow(m_ccm_context.in());
    
  timeout_policy_list[0] = session2context->create_policy ( Messaging::RELATIVE_RT_TIMEOUT_POLICY_TYPE, timeout_any);
    
  /*
   * set policy at object level
   */
  CORBA::Object_var obj_copy =
    obj->_set_policy_overrides ( timeout_policy_list,
                                 CORBA::ADD_OVERRIDE );
    
   timeout_policy_list[0]->destroy();

  PRINT_INFO("Context set_timeout_policy done.");

  return obj_copy._retn();

}
CORBA::Object_ptr
Context::apply_policy(const CORBA::Object_ptr connection,
                      const std::string& comp_oid,
                      const std::string& receptacle_name) 
  throw( CORBA::SystemException )
{

  CORBA::Object_var to_return = CORBA::Object::_duplicate( connection );
  Components::ConfigValues_var config_values;

  PRINT_INFO("Context apply_policy: receptacle name =" << receptacle_name);
         
  /*
   * search the config values matched to the OID
   */
  try{

    CdmwCcmContainer::Session2Context_var session2context = CdmwCcmContainer::Session2Context::_narrow(m_ccm_context.in());
    
    config_values = this->get_config_values(comp_oid);
  
    // Timeout on a receptacle in milliseconds
    unsigned long timeDurationInMs = 0; 

    // default timeout found in the ConfigValues sequence (in milliseconds)
    unsigned long defaultTimeout = 0;
  
    // flag which indicates whether the timeout matched to the receptacle was
    // found or not
    bool receptacle_timeout_found = false;

    // flag which indicates whether the default timeout matched to the recaptacle
    // was found or not
    bool default_timeout_found = false;

    std::string name_config_value;

    /*
     * search the timeout to apply on the receptacle
     */

    // Search in ConfigValues sequence
    for( CORBA::ULong i = 0; i < config_values->length(); i++ ){
      // std::string name_config_value = config_values[i]->name();
      name_config_value = config_values[i]->name();

      if( !name_config_value.compare(CdmwDeployment::USES_PORT_TIMEOUTS )){
        CdmwCcmCif::UsesPortTimeouts_var uses_port_timeouts;
        config_values[i]->value() >>= uses_port_timeouts;

        // Search in UsesPortTimeouts sequence
        for( CORBA::ULong j = 0; j < uses_port_timeouts->length(); j++ ){
          std::string uses_name(uses_port_timeouts[j].uses_name.in());

          if( !uses_name.compare(receptacle_name) ){
            timeDurationInMs = uses_port_timeouts[j].timeout;
            receptacle_timeout_found = true;
            PRINT_INFO("Context apply_policy: timeout of " << receptacle_name
                       << " found. " << "timeout receptacle = "
                       << timeDurationInMs << " (ms)");
            break;
          }else{
            if( !uses_name.compare("") ){
              default_timeout_found = true;
              defaultTimeout = uses_port_timeouts[j].timeout;
            PRINT_INFO("Context apply_policy: default timeout found. "
                       << "default_timeout = " << defaultTimeout << " (ms)");
            }
          }// if, UsesPortTimemout structure
        }// for, UsesPortTimeouts sequence 
        uses_port_timeouts._retn();
        break;
      }//if, ConfigValue valuetype
    }//for, ConfigValues sequence

    if( default_timeout_found == true && receptacle_timeout_found == false ){
      timeDurationInMs = defaultTimeout;
    }

    // Set the policy
    if( receptacle_timeout_found == true || default_timeout_found == true ){

      try{

        to_return = this->set_timeout_policy(timeDurationInMs, to_return.in());

      }
      catch(const CORBA::PolicyError& e){
        PRINT_ERROR("Context apply_policy : set_timeout_policy error! " 
                    << e );
        throw CORBA::INTERNAL(OrbSupport::INTERNALCCMCIFError,
                              CORBA::COMPLETED_NO);
      } 
      catch(const CORBA::SystemException& e){
        PRINT_ERROR("Context apply_policy : set_timeout_policy error! " 
                    << e );
        throw;
      } 
    }
    else{
      PRINT_INFO("Context apply_policy: no default and specific timeout found "
                 << "for the receptacle :" << receptacle_name )
    }
  }
  catch(const Cdmw::Common::NotFoundException& e)
  {
      // no config values matched to the oid
    PRINT_INFO("no config values matched to the oid"
                 << " in the map of the context ( "
                 << "oid = " << comp_oid << "). "
                 << e );
  }  

  PRINT_INFO("Context apply_policy done.");

  return to_return._retn();
  
}

} // namespace CIF 
} // end namespace Cdmw 
} // end namespace CCM 

