/* ===================================================================== */
/*
 * This file is part of CARDAMOM (R) which is jointly developed by THALES
 * and SELEX-SI. It is derivative work based on PERCO Copyright (C) THALES
 * 2000-2003. All rights reserved.
 * Copyright (C) THALES 2004-2006. All rights reserved.
 * 
 * CARDAMOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * CARDAMOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with CARDAMOM; see the file COPYING. If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* ===================================================================== */

// Cdmw Files
#include <Foundation/common/Assert.hpp> 
#include <Foundation/orbsupport/CORBA.hpp>
#include <Foundation/orbsupport/ExceptionMinorCodes.hpp>
#include <Foundation/commonsvcs/naming/NamingInterface.hpp>
#include <Foundation/commonsvcs/serviceregistration/ServiceRegistration_impl.hpp>
#include <Foundation/logging/FunctionLogger.hpp>
#include <FaultTolerance/idllib/CdmwFTReplicationManager.stub.hpp>
#include <FaultTolerance/ftcommon/FTObjectIds.hpp>
#include <FaultTolerance/ftcommon/FTLogger.hpp>


// Local Files
#include <ftultimatefallback/FTUltimateFallbackInterceptor_impl.hpp>


namespace 
{
    const PortableInterceptor::ORBInitInfo::ObjectId FT_SERVER_REQUEST_INTERCEPTOR_NAME 
        = "FTUltimateFallbackInterceptor";

} // end of anonymous namespace


namespace Cdmw
{
    
namespace FT
{
   
namespace UltimateFallback
{

FTUltimateFallbackInterceptor_impl::FTUltimateFallbackInterceptor_impl(
                 PortableInterceptor::ORBInitInfo_ptr info)
{
    CDMW_LOG_FUNCTION(Cdmw::FT::FTLogger::GetLogger(),"");

    //
    // Get the codec factory
    //
    IOP::CodecFactory_var factory = info->codec_factory();
    CDMW_ASSERT(!CORBA::is_nil(factory.in()));

    //
    // Create codec
    //
    IOP::Encoding how;
    how.major_version = 1;
    how.minor_version = 0;
    how.format = IOP::ENCODING_CDR_ENCAPS;

    try
    {
        m_cdrCodec = factory->create_codec(how);
    }
    catch(const IOP::CodecFactory::UnknownEncoding&)
    {
        CDMW_ASSERT(false);
    }

    CDMW_ASSERT(!CORBA::is_nil(m_cdrCodec.in()));
}

char*
FTUltimateFallbackInterceptor_impl::name() 
    throw (CORBA::SystemException)
{
    CDMW_LOG_FUNCTION(Cdmw::FT::FTLogger::GetLogger(),"");
    CDMW_LOG_FUNCTION_RETURN(FT_SERVER_REQUEST_INTERCEPTOR_NAME);
    return CORBA::string_dup(FT_SERVER_REQUEST_INTERCEPTOR_NAME);
}

void
FTUltimateFallbackInterceptor_impl::destroy() 
    throw (CORBA::SystemException)
{
    CDMW_LOG_FUNCTION(Cdmw::FT::FTLogger::GetLogger(),"");
}

void
FTUltimateFallbackInterceptor_impl::receive_request_service_contexts(
        PortableInterceptor::ServerRequestInfo_ptr ri)
    throw (CORBA::SystemException, PortableInterceptor::ForwardRequest)
{
    CDMW_LOG_FUNCTION(Cdmw::FT::FTLogger::GetLogger(),"");

    //
    // Get FT_GROUP_VERSION request service context 
    //
    bool contain_a_ft_group_version_sc = false;
    ::FT::ObjectGroupId object_group_id = 0;
    ::FT::ObjectGroupRefVersion object_group_ref_version = 0;
    
    try
    {
        IOP::ServiceContext_var sc 
            = ri->get_request_service_context(IOP::FT_GROUP_VERSION);

        if (sc->context_id == IOP::FT_GROUP_VERSION)
        {    
            CDMW_DEBUG(Cdmw::FT::FTLogger::GetLogger(), 
                       "Context_id: FT_GROUP_VERSION");
            contain_a_ft_group_version_sc = true;

            CORBA::OctetSeq data(sc->context_data.length());
            data.length(sc->context_data.length());
            for (unsigned int i=0; i < sc->context_data.length(); i++)
            {
                data[i] = sc->context_data[i];
            } 

            //
            // The following code is right but TAO doesn't follow the standard!
            // It replaces the FTGroupVersionServiceContext by the TagFTGroupTaggedComponent.
            //
            //CORBA::Any_var any = m_cdrCodec->decode_value(data, FT::_tc_FTGroupVersionServiceContext);
            //
            //const FT::FTGroupVersionServiceContext* context;
            //if (any.in() >>= context)
            //{
            //    DEBUG_DUMP(location_str, "  FTGroupVersionServiceContext: ");
            //    DEBUG_ECHO("     Object group ref version: " << context->object_group_ref_version);
            //}
            
            CORBA::Any_var any = m_cdrCodec->decode_value(data, ::FT::_tc_TagFTGroupTaggedComponent);
            
            const ::FT::TagFTGroupTaggedComponent* context;
                        
            if (any.in() >>= context)
            {
                CDMW_DEBUG(Cdmw::FT::FTLogger::GetLogger(),
                           "FTGroupVersionServiceContext: "
                           << "\n\t        version = " 
                           << (int)context->component_version.major << "." 
                           << (int)context->component_version.minor
                           << "\n\t        ft_domain_id = " << context->group_domain_id.in()
                           << "\n\t        object_group_id = " << context->object_group_id
                           << "\n\t        object_group_ref_version = " << context->object_group_ref_version);
                
                object_group_id = context->object_group_id;
                object_group_ref_version = context->object_group_ref_version;
            }
            else
            {
                CDMW_ERROR(Cdmw::FT::FTLogger::GetLogger(),
                           "ERROR: Error occured when reading FTGroupVersionServiceContext");
                CORBA::INTERNAL ex(OrbSupport::INTERNALFaultToleranceError, CORBA::COMPLETED_NO);
                CDMW_LOG_FUNCTION_EXCEPTION(ex);
                throw ex;
            }   
        }
        else
        {
            CDMW_ERROR(Cdmw::FT::FTLogger::GetLogger(),
                       "ERROR: Bad context id for IOP::FT_GROUP_VERSION");
            CORBA::INTERNAL ex(OrbSupport::INTERNALFaultToleranceError, CORBA::COMPLETED_NO);
            CDMW_LOG_FUNCTION_EXCEPTION(ex);
            throw ex;
        }
    }
    catch(const CORBA::BAD_PARAM&)
    {
        //DEBUG_DUMP(location_str, "No service context FT_GROUP_VERSION for that request");

        // This is not an error while a request can come to a single reference.
        // with no FT service context.
    }
    catch (const CORBA::SystemException& ex)
    {
        CDMW_ERROR(Cdmw::FT::FTLogger::GetLogger(),
                   "ERROR: get_request_service_context FT_GROUP_VERSION raises exception " << ex );
        CDMW_LOG_FUNCTION_EXCEPTION(ex);
        throw;
    }
    catch (const CORBA::Exception& ex)
    {
        CDMW_ERROR(Cdmw::FT::FTLogger::GetLogger(),
                   "ERROR: get_request_service_context FT_GROUP_VERSION raises CORBA exception " << ex);
        CDMW_LOG_FUNCTION_EXCEPTION(ex);
        throw;
    }
    catch (...)
    {
        CDMW_ERROR(Cdmw::FT::FTLogger::GetLogger(),
                   "ERROR: get_request_service_context raises unknown exception ");
    }

    if (contain_a_ft_group_version_sc)
    {

        CDMW_DEBUG(Cdmw::FT::FTLogger::GetLogger(), 
                   "Fallback called! get the good version with the replication manager");
    
        ::FT::ObjectGroup_var object_group;
        CdmwFT::ReplicationManager_var replication_manager;
        try
        {
            CORBA::Object_ptr obj 
                = Cdmw::CommonSvcs::ServiceRegistration::ServiceRegistration_impl::get_instance(
                      )->get_service(Cdmw::FT::REPLICATION_MANAGER_ObjectId, true);

            replication_manager = CdmwFT::ReplicationManager::_narrow(obj);
    
            if (CORBA::is_nil(replication_manager.in()))
            {
                CDMW_WARN(Cdmw::FT::FTLogger::GetLogger(),
                           "ReplicationManager not accessible for the moment");
                CORBA::TRANSIENT ex(OrbSupport::TRANSIENT, CORBA::COMPLETED_NO);
                CDMW_LOG_FUNCTION_EXCEPTION(ex);
                throw ex;
            }
                    
            // Get the most recent object group reference.
            object_group =
               replication_manager->get_object_group_ref_from_gid(object_group_id);
        }
        catch (const ::FT::ObjectGroupNotFound &)
        {
            CDMW_ERROR(Cdmw::FT::FTLogger::GetLogger(),
                       "ERROR: get_object_group_version_from_gid raises ObjectGroupNotFound exception ");
            CORBA::INTERNAL ex(OrbSupport::INTERNALFaultToleranceError, CORBA::COMPLETED_NO);
            CDMW_LOG_FUNCTION_EXCEPTION(ex);
            throw ex;
        }
        catch (const CORBA::BAD_INV_ORDER& ex)
        {
            CDMW_ERROR(Cdmw::FT::FTLogger::GetLogger(),
                       "ERROR: ServiceRegistration not initialised!\n"
                       << ex );
            CDMW_ASSERT(false);
        }
        catch (const CORBA::SystemException& ex)
        {
            CDMW_ERROR(Cdmw::FT::FTLogger::GetLogger(),
                       "ERROR: System exception raised while getting the group reference version from the replication manager!\n"
                       << ex );
            CDMW_LOG_FUNCTION_EXCEPTION(ex);
            throw;
        }
    
        // Trigger LOCATION_FORWARD_PERM
        CDMW_INFO(Cdmw::FT::FTLogger::GetLogger(),
                  "Ultimate fallback reached for group " << object_group_id
                  << ". Returning latest version of IOGR");
        PortableInterceptor::ForwardRequest ex(object_group.in()); 
        CDMW_LOG_FUNCTION_EXCEPTION(ex);
        throw ex;
    }
}

void
FTUltimateFallbackInterceptor_impl::receive_request(PortableInterceptor::ServerRequestInfo_ptr ri)
        throw (CORBA::SystemException, PortableInterceptor::ForwardRequest)
{
    // nothing to do!
    CDMW_LOG_FUNCTION(Cdmw::FT::FTLogger::GetLogger(),"");
}
    

void
FTUltimateFallbackInterceptor_impl::send_reply(PortableInterceptor::ServerRequestInfo_ptr ri)
        throw (CORBA::SystemException)
{
    // nothing to do!
    CDMW_LOG_FUNCTION(Cdmw::FT::FTLogger::GetLogger(),"");
}

void
FTUltimateFallbackInterceptor_impl::send_other(PortableInterceptor::ServerRequestInfo_ptr ri)
        throw (CORBA::SystemException, PortableInterceptor::ForwardRequest)
{
    // nothing to do!
    CDMW_LOG_FUNCTION(Cdmw::FT::FTLogger::GetLogger(),"");
}

void
FTUltimateFallbackInterceptor_impl::send_exception(PortableInterceptor::ServerRequestInfo_ptr ri)
        throw (CORBA::SystemException, PortableInterceptor::ForwardRequest)
{
    // nothing to do!
    CDMW_LOG_FUNCTION(Cdmw::FT::FTLogger::GetLogger(),"");
}


} // End of namespace UltimateFallback

} // End of namespace FT

} // End of namespace Cdmw


