/* ===================================================================== */
/*
 * This file is part of CARDAMOM (R) which is jointly developed by THALES
 * and SELEX-SI. It is derivative work based on PERCO Copyright (C) THALES 
 * 2000-2003. All rights reserved.
 * Copyright (C) THALES 2004-2005. All rights reserved.
 * Copyright (C) SELEX-SI and THALES 2006. All rights reserved.
 * 
 * CARDAMOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * CARDAMOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with CARDAMOM; see the file COPYING. If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* ===================================================================== */


#include <Foundation/commonsvcs/serviceregistration/ServiceRegistration_impl.hpp>
#include <Foundation/ossupport/OS.hpp>
#include <Foundation/orbsupport/OrbSupport.hpp>
#include <Foundation/osthreads/MutexGuard.hpp>
#include <Foundation/orbsupport/ExceptionMinorCodes.hpp>
#include <Foundation/logging/Logger.hpp>
#include <Foundation/logging/FunctionLogger.hpp>
#include <Foundation/commonsvcs/time/Util.hpp>

#include "serviceregistration/NamingContext_impl.hpp"

namespace
{
   const char* const SVC_REG_LOGGER_NAME = "Cdmw.FDS.SVC_REG";
   Cdmw::Logging::Logger *
      GetLogger()
      {
         static Cdmw::Logging::Logger * logger =
              Cdmw::Logging::Logger::Get_Logger(SVC_REG_LOGGER_NAME );
         return logger;

      }
}

namespace Cdmw
{
namespace CommonSvcs
{
namespace ServiceRegistration
{
    ServiceRegistration_impl* ServiceRegistration_impl::m_instance = NULL;

    ServiceRegistration_impl* ServiceRegistration_impl::init(CORBA::ORB_ptr orb,
                                                             PortableGroup::GOA_ptr goa,
                                                             std::string service_locator_address,
                                                             std::string registration_service_address,
                                                             int nb_retries,
                                                             int time_interval)
        throw (CORBA::SystemException)
    {
        CDMW_LOG_FUNCTION(::GetLogger(),"registration_service_address="
             << registration_service_address << ",nb_retries="
             << nb_retries << ",time_interval="<<time_interval);

        if (m_instance)
        {
            CORBA::BAD_INV_ORDER ex (Cdmw::OrbSupport::BAD_INV_ORDERInitAlreadyDone,
                                 CORBA::COMPLETED_NO);
            throw ex;
        }
        else
        {
            m_instance = new ServiceRegistration_impl(orb,
                                                      goa,
                                                      service_locator_address,
                                                      registration_service_address,
                                                      nb_retries,
                                                      time_interval);

            NamingContext_impl* nc_impl = new NamingContext_impl(m_instance);
            CosNaming::NamingContext_var nc = nc_impl->_this();
            
            // publishes the daemon naming context corbaloc
            Cdmw::OrbSupport::OrbSupport::bind_object_to_corbaloc
                (orb, "NameService", nc.in() );
    
            orb->register_initial_reference("NameService", nc.in());  
        }

        return m_instance;
    }

    ServiceRegistration_impl* ServiceRegistration_impl::get_instance() 
        throw (CORBA::SystemException)
    {
        if (m_instance)
        {
            return m_instance;
        }
        else
        {
            CORBA::BAD_INV_ORDER ex (Cdmw::OrbSupport::BAD_INV_ORDERInitNotCalled,
                                     CORBA::COMPLETED_NO);
            throw ex;
        }
    }

    ServiceRegistration_impl::ServiceRegistration_impl(CORBA::ORB_ptr orb,
                                                       PortableGroup::GOA_ptr goa,
                                                       std::string service_locator_address,
                                                       std::string registration_service_address,
                                                       int nb_retries,
                                                       int time_interval)
        : m_servicesCondition(m_servicesMutex),
          m_orb(CORBA::ORB::_duplicate(orb)),
          m_goa(PortableGroup::GOA::_duplicate(goa)),
          m_nb_retries(nb_retries),
          m_time_interval(time_interval),
          m_registration_service_address(registration_service_address)
    {
        CDMW_LOG_FUNCTION(::GetLogger(),"");
        
        // get multicast object group of service locator to interogate
        CORBA::Object_var service_locator_group = 
                m_orb->string_to_object(service_locator_address.c_str());
        
        m_service_locator = 
            CdmwServiceLocator::ServiceLocator::_unchecked_narrow (service_locator_group.in());
        
        // get multicast object group of registration service 
        CORBA::Object_var registration_service_group = 
                m_orb->string_to_object(registration_service_address.c_str());
        
        // get an object id on this multicast object group
        PortableServer::ObjectId_var registration_id = 
                m_goa->create_id_for_reference(registration_service_group.in());
        
        // associate this servant to the multicast object id
        m_goa->activate_object_with_id (registration_id.in(), this);         
    }
    
    
    ServiceRegistration_impl::~ServiceRegistration_impl()
    {
    }
    

    //
    // Register the service. Called by the service locator
    // NOTA: local cache is always updated no mater how 'old' the received reference is!
    // (there is no versioning for the reference)
    //
    void ServiceRegistration_impl::register_service(const char* name, CORBA::Object_ptr ref)
        throw(CORBA::SystemException)
    {
        CDMW_LOG_FUNCTION(::GetLogger(),"name="<<name);
        CDMW_MUTEX_GUARD(m_servicesMutex);

        // store the service reference in the map
        //    CORBA::Object_var obj = CORBA::Object::_duplicate(ref);
        m_services[name] = CacheData(ref);

          // broadcast the condition
        m_servicesCondition.broadcast();
                
//         // first search the service in map
//         ServiceMap::iterator iter = m_services.find(name);
        
//         // if not found
//         if (iter == m_services.end())
//         {
//              // store the service reference in the map
//              CORBA::Object_var obj = CORBA::Object::_duplicate(ref);
//              m_services[name] = obj;
             
//              // broadcast the condition
//              m_servicesCondition.broadcast();
//         } 
    }
    
    //
    // Register the service. Called locally
    // NOTA: local cache is always updated no mater how 'old' the received reference is!
    // (there is no versioning for the reference)
    //
    void ServiceRegistration_impl::register_local_service(const char* name, CORBA::Object_ptr ref)
    {
        CDMW_LOG_FUNCTION(::GetLogger(),"name="<<name);

        CDMW_MUTEX_GUARD(m_servicesMutex);
        // store the service reference in the map
        m_services[name] = CacheData(ref);

                
//         // first search the service in map
//         ServiceMap::iterator iter = m_services.find(name);
        
//         // if not found
//         if (iter == m_services.end())
//         {
//              // store the service reference in the map
//              CORBA::Object_var obj = CORBA::Object::_duplicate(ref);
//              m_services[name] = obj;
//         } 
    }
    

    //
    // get the registered service
    // if not already registered, discover by interogating the service locator
    //
    CORBA::Object_ptr ServiceRegistration_impl::get_service(const char * name, bool do_refresh)
    {
        CDMW_LOG_FUNCTION(::GetLogger(),"name="<<name<<",refresh="<<do_refresh);

        CORBA::Object_var service_ref = CORBA::Object::_nil();
        
        // lock mutex of service map
        CDMW_MUTEX_GUARD(m_servicesMutex);
        
        // first search the servive in map
        InternalServiceMap::iterator iter = m_services.find(name);
        
        // if found
        // Do not return old value if refresh is requested and no new update has arrived!
        if (iter != m_services.end()) {
            if (do_refresh)
                iter->second.m_is_refreshed=false;
            else
            {
                service_ref = (*iter).second.m_obj;
                CDMW_LOG_FUNCTION_RETURN("");
                return service_ref._retn();
            }
        }
        
        // the service is not in the map, interogate the service locator
        // in multicast

        // calculate the max time out to wait response
        //unsigned long max_timeout = (unsigned long) (m_nb_retries * m_time_interval);
        std::size_t max_timeout = static_cast<std::size_t>(m_time_interval) * m_nb_retries;
        
        // first get the start time
        // may throw ::Cdmw::InternalErrorException TODO!!!
        //OsSupport::OS::Timeval start_time = OsSupport::OS::get_time();
        TimeBase::TimeT start_time = Time::Util::get_current_time();
        
        // call the service locator in multicast
        CDMW_DEBUG(::GetLogger(),"locate_service("<<name<<")");

        m_service_locator->locate_service(name, m_registration_service_address.c_str());
        
        bool timeout = false;

        //unsigned long allowed_timeout = (unsigned long) m_time_interval;
        std::size_t allowed_timeout = static_cast<std::size_t>(m_time_interval);
        
        // wait the response
        while (!timeout)
        {
            // get the wait start time
//            OsSupport::OS::Timeval wait_start_time = OsSupport::OS::get_time();
            TimeBase::TimeT wait_start_time = Time::Util::get_current_time();
        
            // wait the condition
            CDMW_DEBUG(::GetLogger(),"locate_service("<<name<<") wait the condition");
            OsSupport::Condition::WaitReturn statusWait =
                m_servicesCondition.wait(allowed_timeout);
            
            // condition wakeup
            // search the service in map
            InternalServiceMap::iterator iter = m_services.find(name);
    
            // if service reference received
            if (iter != m_services.end())
            {
                if (!do_refresh || (*iter).second.m_is_refreshed)
                {
                   CDMW_DEBUG(::GetLogger(),"locate_service("<<name<<") service reference received");
                   service_ref = (*iter).second.m_obj;
                   break;
                }
            }
            
            // if time out
            if (statusWait == OsSupport::Condition::TIMEDOUT)
            {       
                // get the current time
                //OsSupport::OS::Timeval time = OsSupport::OS::get_time();
                TimeBase::TimeT now = Time::Util::get_current_time();
                     
                // Compute the remaining timeout (in ms)
                TimeBase::TimeT time_diff = now - start_time; // This is in 10^-7 s

                std::size_t elapsed_time = static_cast<std::size_t>(time_diff / 10000);

//                 unsigned long elapsed_time =
//                    (time.seconds - start_time.seconds) * 1000000;
//                 elapsed_time += (time.microseconds - start_time.microseconds);
//                 elapsed_time = elapsed_time / 1000;
                
                // timeout has not been reached, retry a new interogation
                if (elapsed_time < max_timeout)
                {
                    // calculate the allowed timeout to wait a new response
                    // limited to time interval
                    allowed_timeout = max_timeout - elapsed_time;

                    if (allowed_timeout > static_cast<std::size_t>(m_time_interval))

                    {
                       allowed_timeout = static_cast<std::size_t>(m_time_interval);
                    }
                    
                    // recall the service locator in multicast
                   CDMW_DEBUG(::GetLogger(),"locate_service("<<name<<") after timeout");

                    m_service_locator->locate_service(
                             name, m_registration_service_address.c_str());
                }
                // else time out has been reached, service is not found
                else
                {
                    CDMW_DEBUG(::GetLogger(),"locate_service("<<name<<") time out has been reached, service is not found");
                    timeout = true;
                }                    
            }  
            
            // condition is wakeup but not by broadcast or timeout
            // condition must return to wait
            else
            {
                // get the current time
                //OsSupport::OS::Timeval time = OsSupport::OS::get_time();
                TimeBase::TimeT now = Time::Util::get_current_time();
                
                // Compute the remaining timeout for waiting (in ms)
                TimeBase::TimeT time_diff = now - wait_start_time; // This is in 10^-7 s
                std::size_t elapsed_time = static_cast<std::size_t>(time_diff / 10000);
//                 unsigned long elapsed_time =
//                    (time.seconds - wait_start_time.seconds) * 1000000;
//                 elapsed_time += (time.microseconds - wait_start_time.microseconds);
//                 elapsed_time = elapsed_time / 1000;
                
                // calculate the remaining allowed timeout to wait response
                if (elapsed_time >= allowed_timeout)
                {
                    allowed_timeout = 0;
                }
                else
                {
                    allowed_timeout = allowed_timeout - elapsed_time;
                }
            }                        
        }

        if (::GetLogger()->is_debug_enabled()) {
           CORBA::String_var s__ = m_orb->object_to_string(service_ref.in());
           CDMW_DEBUG(::GetLogger(),"locate_service("<<name<<"): Returning  "  << "\n"
                      << "\t IOR =" << s__.in() );

        }
        
        return service_ref._retn();
    }

   
    //
    // Returns a copy of the underlying service map.
    //
    ServiceRegistration_impl::ServiceMap ServiceRegistration_impl::get_known_services()
    {
        CDMW_LOG_FUNCTION(::GetLogger(),"");
        ServiceMap result;
        for (InternalServiceMap::const_iterator it = m_services.begin();
             it != m_services.end();
             ++it) {
             result[it->first]=it->second.m_obj;
        }
        return result;
    }
    
} // End namespace ServiceRegistration
} // End namespace CommonSvcs
} // End namespace Cdmw
