/* ===================================================================== */
/*
 * This file is part of CARDAMOM (R) which is jointly developed by THALES
 * and SELEX-SI. It is derivative work based on PERCO Copyright (C) THALES
 * 2000-2003. All rights reserved.
 * Copyright (C) THALES 2004-2005. All rights reserved.
 * Copyright (C) SELEX-SI and THALES 2006. All rights reserved.
 *
 * CARDAMOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CARDAMOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with CARDAMOM; see the file COPYING. If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* ===================================================================== */
package cdmw.logging;

public class LogManager
{
    private static LogManager M_Instance;   
    private LogManagerDelegate m_defaultDelegate;
    private LogManagerDelegate m_alternateDelegate;
    private FilterMngr m_filterManager;
    private LevelType level;
    private LogLevel logLevel;

    /**
     * Initialise the LogManager.
     *
     * It creates a default delegate, a target to the logs.
     *
     * @return the LogManager that has been initialised.
     */

    protected LogManager() {
        m_defaultDelegate = null;
        m_alternateDelegate = null;
        m_filterManager = null;
    }

    /**
     * Initialise the LogManager.
     *
     * It creates a default delegate, a target to the logs.
     * Command-line arguments will be used to initialise the
     * default logger.
     *
     * @return the LogManager that has been initialised.
     */

    protected LogManager(String[] args) {
        m_defaultDelegate = null;
        m_alternateDelegate = null;
        logLevel = new LogLevel();
        level = new LevelType();
        // create the filter manager with default levels
        level.setType(logLevel.INFO|logLevel.WARN|logLevel.ERROR|logLevel.FATAL);
        m_filterManager = new FilterMngr(level.getType());

        // create the default delegate of LogManager
        m_defaultDelegate = new LogManagerDefaultDelegate( args, m_filterManager);
    }

    /**
     * Initialise the LogManager.
     */
    public static synchronized LogManager Init(String[] args) {
        if (M_Instance == null)
        { 
            M_Instance = new LogManager(args);
        }

        return M_Instance;
    }

    /**
     * Send a message to the logger.
     *
     * @param header the message header.
     * @param msg the message.
     *
     * @return true if successful.
     */

    public synchronized boolean addLog(LogHeader header,String msg) {
        if (m_alternateDelegate != null) 
        {
            return m_alternateDelegate.addLog(header, msg);
        }
        else 
        {
            return m_defaultDelegate.addLog(header, msg);
        }
    }

    /**
     * Register an alternate delegate.
     *
     * @param logmgr the alternate delegate.
     *
     * @return the previous registered alternate delegate.
     */

    public synchronized LogManagerDelegate registerAlternateDelegate(LogManagerDelegate alt) {
        LogManagerDelegate old = m_alternateDelegate;
        m_alternateDelegate = alt;

        return old;
    } 

    /**
     * Unregister the alternate delegate.
     *
     * @param logmgr the alternate delegate.
     *
     * @return true if the alternate delegate was previously registered.
     */

    public synchronized boolean unregisterAlternateDelegate(LogManagerDelegate alt) {

        boolean result = false;

        if (alt == m_alternateDelegate) 
        {
            result = true;
            m_alternateDelegate = null;
        }

        return result;
     }

    /**
     * Get the LogManager.
     *
     * @return the LogManager Instance.
     */

    public static LogManager Instance() {
        return M_Instance;
     }

    /**
     * Get the alternate delegate.
     *
     * @return the alternate delegate.
     */
    public LogManagerDelegate getAlternateDelegate() {
        return m_alternateDelegate;
    }

    /**
     * Get the filter manager.
     *
     * @return the filter manager.
     */
    public FilterMngr getFilterManager() {
        return m_filterManager;
    }
}
