/* ===================================================================== */
/*
 * This file is part of CARDAMOM (R) which is jointly developed by THALES
 * and SELEX-SI. It is derivative work based on PERCO Copyright (C) THALES
 * 2000-2003. All rights reserved.
 * Copyright (C) THALES 2004-2005. All rights reserved.
 * Copyright (C) SELEX-SI and THALES 2006. All rights reserved.
 *
 * CARDAMOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CARDAMOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with CARDAMOM; see the file COPYING. If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* ===================================================================== */
package cdmw.logging;
import java.lang.Thread;
import java.util.*;
import cdmw.ossupport.*;

public class Logger
{

    /**
     * Default LevelTypes object
     */
    private static final LevelType FATAL   = new LevelType(LogLevel.FATAL  );
    private static final LevelType ERROR   = new LevelType(LogLevel.ERROR  );
    private static final LevelType WARN    = new LevelType(LogLevel.WARN   );
    private static final LevelType INFO    = new LevelType(LogLevel.INFO   );
    private static final LevelType DEBUG   = new LevelType(LogLevel.DEBUG  );
    private static final LevelType DEBUG_1 = new LevelType(LogLevel.DEBUG_1);
    private static final LevelType DEBUG_2 = new LevelType(LogLevel.DEBUG_2);
    private static final LevelType DEBUG_3 = new LevelType(LogLevel.DEBUG_3);
    private static final LevelType DEBUG_4 = new LevelType(LogLevel.DEBUG_4);
    private static final LevelType DEBUG_5 = new LevelType(LogLevel.DEBUG_5);
    private static final LevelType DEBUG_6 = new LevelType(LogLevel.DEBUG_6);
    private static final LevelType DEBUG_7 = new LevelType(LogLevel.DEBUG_7);
    private static final LevelType DEBUG_8 = new LevelType(LogLevel.DEBUG_8);
    private static final LevelType DEBUG_9 = new LevelType(LogLevel.DEBUG_9);

    
    private Filter m_filter = new Filter();
    private String m_domain;
    private String m_componentName;
    private LogManager m_logMngr = new LogManager();
   
    /**
     * Create a Logger.
     *
     */
    protected Logger() {
    }

    /**
     * Create a Logger .
     *
     * @param logMngr the LogManager.
     * @param componentName the name of the component.
     * @param domain the name of the domain.
     */
    protected Logger(LogManager logMngr, String componentName, String domain) {
        m_logMngr = logMngr;
        m_componentName = componentName;
        m_domain = domain;
        FilterMngr filterMngr = logMngr.getFilterManager();
        m_filter = filterMngr.getFilterRef(componentName,domain);
    }
    
    /**
     * Create a Logger.
     *
     * @param logMngr the LogManager.
     * @param domain the name of the domain.
     */
    protected Logger(LogManager logMngr, String domain) {
        m_logMngr = logMngr;
        m_componentName = "";
        m_domain = domain;
        FilterMngr filterMngr = logMngr.getFilterManager();
        m_filter = filterMngr.getFilterRef(domain);
    }

    
    /**
     * Log a generic message.
     *
     * @param level the message level.
     * @param message the message.
     * @param file the file from which the message is output.
     * @param line the line number in the file.
     */

    public void log (LevelType level, String message, Throwable thr) {
        LogHeader header = new LogHeader();
        Thread t = Thread.currentThread();
        header.m_fileLine = 0;
        header.m_fileName = "";
        header.m_userLevel =level;
        header.m_userDomain = this.m_domain;
        header.m_componentName = this.m_componentName;
        header.m_tid = t.getId();
        header.m_timestamping = new cdmw.ossupport.Timeval(Calendar.getInstance().getTimeInMillis());

       if (thr != null) {
           StackTraceElement[] stack = thr.getStackTrace();
           if (stack[0].getFileName() != null) {
               header.m_fileName = stack[0].getFileName();
               header.m_fileLine = stack[0].getLineNumber();
          } else {
               header.m_fileName = "no_debug_info";
          }
       }


        m_logMngr.addLog(header, message); 

    }

    /**
     * Find the logger associated to the specified domain name.
     *
     * NOTA: the caller must assume the ownership of the returned                 * Logger object.
     *
     * @param domain the domain name.
     *
     * @return the logger if found, NULL otherwise.
     */

    public static Logger Get_Logger(String domain) {
        LogManager logMngr = LogManager.Instance();
        return new Logger(logMngr, "", domain);
    }

    /**
     * Find the logger associated to the specified component name and
     * domain name.
     *
     * NOTA: the caller must assume the ownership of the returned                 * Logger object.
     *
     * @param component the component name.
     * @param domain the domain name.
     *
     * @return the logger if found, NULL otherwise.
     */

    public static Logger Get_Logger(String componentName, String domain) {
        LogManager logMngr = LogManager.Instance();
        return new Logger(logMngr, componentName, domain);
    }

    /**
     * Log a debug message.
     *
     * @param message the message.
     * @param file the file from which the message is output.
     * @param line the line number in the file.
     */

    public void debug (String message, Throwable thr) {
        log(DEBUG, message, thr);
    }

    /**
     * Log a info message.
     *
     * @param message the message.
     * @param file the file from which the message is output.
     * @param line the line number in the file.
     */

    public void info (String message, Throwable thr) {
        log(INFO, message, thr);
    }

    
    /**
     * Log a warn message.
     *
     * @param message the message.
     * @param file the file from which the message is output.
     * @param line the line number in the file.
     */

    public void warn (String message, Throwable thr) {
        log(WARN, message, thr);
    }
    
    /**
     * Log a error message.
     *
     * @param message the message.
     * @param file the file from which the message is output.
     * @param line the line number in the file.
     */

    public void error (String message, Throwable thr) {
        log(ERROR, message, thr);
    }
    

    /**
     * Log a fatal message.
     *
     * @param message the message.
     * @param file the file from which the message is output.
     * @param line the line number in the file.
     */

    public void fatal (String message, Throwable thr) {
        log(FATAL, message, thr);
    }

    /**
     * Check whether a log level is enabled.
     *
     * @param line the log level
     *
     * @return true if enabled.
     */

    public boolean isEnabledFor(LevelType level) {
        return m_filter.isLevelActivated(level.getType());
    }

    
    /**
     * Check whether DEBUG level is enabled.
     *
     * @param line the log level
     *
     * @return true if enabled.
     */
    public boolean isDebugEnabled() {
        return m_filter.isLevelActivated(LogLevel.DEBUG);
    }

    /**
     * Check whether DEBUG level is enabled.
     *
     * @param line the log level
     *
     * @return true if enabled.
     */

    public boolean isInfoEnabled() {
        return m_filter.isLevelActivated(LogLevel.INFO);
    }

    /**
     * Check whether WARN level is enabled.
     *
     * @param line the log level
     *
     * @return true if enabled.
     */

    public boolean isWarnEnabled() {
        return m_filter.isLevelActivated(LogLevel.WARN);
    }

    /**
     * Check whether ERROR level is enabled.
     *
     * @param line the log level
     *
     * @return true if enabled.
     */

    public boolean isErrorEnabled() {
        return m_filter.isLevelActivated(LogLevel.ERROR);
    }

    
    /**
     * Check whether FATAL level is enabled.
     *
     * @param line the log level
     *
     * @return true if enabled.
     */

    public boolean isFatalEnabled() {
        return m_filter.isLevelActivated(LogLevel.FATAL);
    }


    static public void CDMW_LOG(Logger logger, LevelType level, String msg, Throwable thr) {
        if (logger.isEnabledFor(level)) {
            logger.log(level, msg, thr);
        }
    }

            
    static public void CDMW_FATAL(Logger logger, String msg, Throwable thr) {
        if (logger.isFatalEnabled()) {
            logger.log(FATAL, msg, thr);
        }
    }

    static public void CDMW_ERROR(Logger logger, String msg, Throwable thr) {
        if (logger.isErrorEnabled()) {
            logger.log(ERROR, msg, thr);
        }
    }
    
    static public void CDMW_WARN(Logger logger, String msg, Throwable thr) {
        if (logger.isWarnEnabled()) {
            logger.log(WARN, msg, thr);
        }
    }
 
    static public void CDMW_INFO(Logger logger, String msg, Throwable thr) {
        if (logger.isInfoEnabled()) {
            logger.log(INFO, msg, thr);
        }
    }

    static public void CDMW_DEBUG(Logger logger, String msg, Throwable thr) {
        if (logger.isDebugEnabled()) {
            logger.log(DEBUG, msg, thr);
        }
    }
    
    static public void CDMW_DEBUG_1(Logger logger, String msg, Throwable thr) {
        if (logger.isEnabledFor(DEBUG_1)) {
            logger.log(DEBUG_1, msg, thr);
        }
    }
    
    static public void CDMW_DEBUG_2(Logger logger, String msg, Throwable thr) {
        if (logger.isEnabledFor(DEBUG_2)) {
            logger.log(DEBUG_2, msg, thr);
        }
    }
    
    static public void CDMW_DEBUG_3(Logger logger, String msg, Throwable thr) {
        if (logger.isEnabledFor(DEBUG_3)) {
            logger.log(DEBUG_3, msg, thr);
        }
    }
    
    static public void CDMW_DEBUG_4(Logger logger, String msg, Throwable thr) {
        if (logger.isEnabledFor(DEBUG_4)) {
            logger.log(DEBUG_4, msg, thr);
        }
    }
    
   static public void CDMW_DEBUG_5(Logger logger, String msg, Throwable thr) {
        if (logger.isEnabledFor(DEBUG_5)) {
            logger.log(DEBUG_5, msg, thr);
        }
    }
    
    static public void CDMW_DEBUG_6(Logger logger, String msg, Throwable thr) {
        if (logger.isEnabledFor(DEBUG_6)) {
            logger.log(DEBUG_6, msg, thr);
        }
    }
    
    static public void CDMW_DEBUG_7(Logger logger, String msg, Throwable thr) {
        if (logger.isEnabledFor(DEBUG_7)) {
            logger.log(DEBUG_7, msg, thr);
        }
    }
    
    static public void CDMW_DEBUG_8(Logger logger, String msg, Throwable thr) {
        if (logger.isEnabledFor(DEBUG_8)) {
            logger.log(DEBUG_8, msg, thr);
        }
    }
    
    static public void CDMW_DEBUG_9(Logger logger, String msg, Throwable thr) {
        if (logger.isEnabledFor(DEBUG_9)) {
            logger.log(DEBUG_9, msg, thr);
        }
    }
}
