/* ===================================================================== */
/*
 * This file is part of CARDAMOM (R) which is jointly developed by THALES
 * and SELEX-SI. It is derivative work based on PERCO Copyright (C) THALES 
 * 2000-2003. All rights reserved.
 * Copyright (C) THALES 2004-2005. All rights reserved.
 * Copyright (C) SELEX-SI and THALES 2006. All rights reserved.
 * 
 * CARDAMOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * CARDAMOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with CARDAMOM; see the file COPYING. If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* ===================================================================== */

#include "test_timer_invocation/ClientProcessControl.hpp"

namespace TimerInvocation
{

ClientProcessControl::ClientProcessControl ( CORBA::ORB_ptr orb, utils::cfgData& cfg )
throw( CORBA::SystemException )
        : m_orb ( CORBA::ORB::_duplicate( orb ) ),
        m_applicationName ( "" ),
        m_processName ( "" ),
        m_clock( 0 ),
        m_controller( 0 ),
        m_periodic( 0 ),
        m_cfg( cfg )
{
    // ===================================================
    // Instantiate time reading object
    // ===================================================
    std::string measure_str = cfg.measure_type[ 0 ];

    if ( measure_str.compare( MEASURE_TH ) == 0 ) {
        m_measure = new ClockCpuTime;
    }
    else if ( measure_str.compare( MEASURE_TK ) == 0 ) {
        m_measure = new ClockCpuTick;
    }
    else // using default
    {
        m_measure = new ClockHighResTime;
    }

    m_niter = m_cfg.niter[ 0 ];
    m_fileprefix = m_cfg.fileprefix[ 0 ];
}

ClientProcessControl::~ClientProcessControl()
throw()
{}


// process initialisation called by platformmngt
void ClientProcessControl::on_initialise (
    const CdmwPlatformMngtBase::StartupKind& startup_kind )
throw( CORBA::SystemException )
{


    // ===================================================
    // Get the process name and application name
    // ===================================================
    m_processName = Cdmw::CdmwInit::InitUtils::get_cdmw_process_name();

    m_applicationName = Cdmw::CdmwInit::InitUtils::get_cdmw_application_name();

    // =================================================================
    // example of using the PlatformInterface for notifying a message
    // =================================================================
    Cdmw::PlatformMngt::PlatformInterface
    ::Notify_message( CdmwPlatformMngtBase::INF,
                      m_processName.c_str(),
                      ">>>>>>>>>>>>>> Initialisation requested by supervision" );

    // ======================
    // retrieve LocalExecutor
    // ======================
    CORBA::Object_var obj =
        m_orb->resolve_initial_references( "ClockService" );

    if ( CORBA::is_nil( obj.in() ) ) {
        std::cerr << "Client:>> Unable to resolve initial reference"
        << std::endl;
        exit( 1 );
    }

    //=====================================================================
    CosClockService::ClockCatalog_var clock_catalog =
        CosClockService::ClockCatalog::_narrow( obj.in() );

    if ( CORBA::is_nil( clock_catalog.in() ) ) {
        std::cerr << "Client:>> Unable to resolve ClockCatalog"
        << std::endl;
        exit( 1 );
    }

    //=====================================================================
    CosClockService::ClockCatalog::ClockEntry_var
    entry = clock_catalog->get_entry( "LocalExecutor" );

    m_clock = entry->subject;

    if ( CORBA::is_nil( m_clock.in() ) ) {
        std::cerr << "Client:>> Unable to resolve LocalExecutor"
        << std::endl;
        exit( 1 );
    }
} // on_initialise()

///////////////////////////////////////////////////////////////////////////////

// process to run called by platformmngt
void ClientProcessControl::on_run()
throw( CdmwPlatformMngt::ProcessDelegate::NotReadyToRun, CORBA::SystemException )
{

    // =================================================================
    // example of using the PlatformInterface for notifying a message
    // =================================================================
    Cdmw::PlatformMngt::PlatformInterface
    ::Notify_message( CdmwPlatformMngtBase::INF,
                      m_processName.c_str(),
                      ">>>>>>>>>>>>>> Run requested by supervision" );


    //-----------------------------------------------------------------------------
    CosClockService::PeriodicExecution::Executor_var executor
    = CosClockService::PeriodicExecution::Executor::_narrow( m_clock.in() );

    remove( TEST_EOF ); // unlink file

    for ( size_t i_per = 0; i_per < m_cfg.delay.size(); i_per++ ) {

        remove( PERF_EOF ); // unlink file

        m_period = static_cast<TimeBase::TimeT>( m_cfg.delay[ i_per ] );

        std::cout << "==> Processing period = " << m_period << " ms" << std::endl;

        // transform millisec in TimeBase::TimeT formats
        // Note: 1s == 10000000 TimeBase::TimeT units.
        m_period = m_period * 10000;

        m_periodic = new perfPeriodic::myPeriodic( m_niter );

    CosClockService::PeriodicExecution::Periodic_var
    periodic = m_periodic->_this();

    m_controller = executor->enable_periodic_execution( periodic.in() );

        CORBA::Any aPar;
        m_controller->start( m_period,
                             0,
                             m_niter,
                             aPar );
    //-----------------------------------------------------------------------------
    ifstream fin;

    fin.open( PERF_EOF );

    while ( !fin.is_open() ) {
#ifdef DEBUG_ON
        std::cerr << "ClientProcessControl::on_run(): Waiting for Timer Test..\n";
#endif

        sleep( 1 );
        fin.open( PERF_EOF );
    }

    fin.close();
        std::ostringstream nameStream;
        nameStream << m_cfg.inv_type[ 0 ] << "_" << m_period << "_" << "DELTA_C.dat";
        std::string filedeltaname ( m_fileprefix + nameStream.str() );


        m_periodic->writeRes( m_period, filedeltaname );
        delete m_periodic;
    }

    //-----------------------------------------------------------------------------
    // write file for shell script execution control
    ofstream feof ( TEST_EOF );

    feof << "EOF" << std::endl;

    feof.close();

    std::cerr << "** test execution successfull terminated: Please STOP the system!" << std::endl;
} // on_run()

///////////////////////////////////////////////////////////////////////////////

// process to stop called by platformmngt
void ClientProcessControl::on_stop()
throw( CORBA::SystemException )
{

    // =================================================================
    // example of using the PlatformInterface for notifying a message
    // =================================================================
    Cdmw::PlatformMngt::PlatformInterface
    ::Notify_message( CdmwPlatformMngtBase::INF,
                      m_processName.c_str(),
                      ">>>>>>>>>>>>>> Stop requested by supervision" );
    m_controller->terminate();
    m_orb->shutdown( true );
} // on_stop()

}

; // End namespace TimerInvocation
