/* ===================================================================== */
/*
 * This file is part of CARDAMOM (R) which is jointly developed by THALES
 * and SELEX-SI. It is derivative work based on PERCO Copyright (C) THALES
 * 2000-2003. All rights reserved.
 * Copyright (C) THALES 2004-2005. All rights reserved.
 * Copyright (C) SELEX-SI and THALES 2006. All rights reserved.
 * 
 * CARDAMOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * CARDAMOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with CARDAMOM; see the file COPYING. If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* ===================================================================== */


#include <limits>
#include <iostream>
#include "platformsystemlibrary/Host_impl.hpp"
#include "platformsystemlibrary/System_impl.hpp"
#include "platformsystemlibrary/SystemLogMessageId.hpp"

#include "SystemMngt/platformlibrary/Configuration.hpp"
#include "SystemMngt/platformlibrary/EventHeaderFactory.hpp"
#include "SystemMngt/platformlibrary/LogMngr.hpp"
#include "SystemMngt/platformlibrary/SupervisionEventNotifier.hpp"
#include "SystemMngt/platformlibrary/EventNotifierManager.hpp"

#include "SystemMngt/platformelementlibrary/EntityContainer_impl.hpp"
#include "SystemMngt/platformelementlibrary/RWEntityContainer_impl.hpp"
#include "SystemMngt/platformelementlibrary/EntityContainer.hpp"
#include "SystemMngt/platformvaluetypes/EntityStatusChange_impl.hpp"
#include "SystemMngt/platformvaluetypes/HostStatusChange_impl.hpp"
#include "SystemMngt/platforminterface/ServiceNames.hpp"
#include "platformapplicationlibrary/ProcessesIterator_impl.hpp"
#include "platformapplicationlibrary/Application_impl.hpp"
#include "platformapplicationlibrary/Process_impl.hpp"
#include "idllib/CdmwPlatformMngtSystem.stub.hpp"
#include "idllib/CdmwPlatformMngtApplication.stub.hpp"
#include "idllib/CdmwPlatformMngtProcess.stub.hpp"
#include "SystemMngt/tools/ManagedElementKey.hpp"
#include <Foundation/common/Locations.hpp>
#include "Foundation/orbsupport/ExceptionMinorCodes.hpp"
#include "Foundation/osthreads/MutexGuard.hpp"
#include "Foundation/logging/Logger.hpp"
#include "Repository/repositoryinterface/RepositoryInterface.hpp"

using namespace std;

namespace Cdmw
{
namespace PlatformMngt
{
        
class Host_impl;
class System_impl;
        
        
        
// ----------------------------------------------------------------------
// HostRespondingCallback class.
// ----------------------------------------------------------------------
HostRespondingCallback::HostRespondingCallback (Host_impl* host)
            : Callback()
{
            m_host = host;
}
        
        
HostRespondingCallback::~HostRespondingCallback ()
{
}
        
        
void HostRespondingCallback::execute()
            throw()
{
            CdmwPlatformMngtBase::TimeStamp time_stamp =
                PlatformMngt::EventHeaderFactory::createTimeStamp();
            m_host->setStatus (CdmwPlatformMngt::HOST_RESPONDING, time_stamp);

            
}
        
// ----------------------------------------------------------------------
// HostNotRespondingCallback class.
// ----------------------------------------------------------------------
HostNotRespondingCallback::HostNotRespondingCallback (Host_impl* host)
            : Callback()
{
            m_host = host;
}
        
        
HostNotRespondingCallback::~HostNotRespondingCallback ()
{
}
        
        
void HostNotRespondingCallback::execute()
            throw()
{
            CdmwPlatformMngtBase::TimeStamp time_stamp =
                PlatformMngt::EventHeaderFactory::createTimeStamp();
            m_host->setStatus (CdmwPlatformMngt::HOST_UNREACHABLE, time_stamp);
            
}
        
        
        
        
// ----------------------------------------------------------------------
// Host_impl class.
// ----------------------------------------------------------------------
        
        
#ifdef CDMW_USE_FAULTTOLERANCE
Host_impl::Host_impl (PortableServer::POA_ptr poa,
            const CdmwPlatformMngt::HostDef& host_def,
            const char* host_path,
            System_impl* system,
            CdmwPlatformMngt::RWElementRepository_ptr element_repository,                      
            CdmwFT::FaultManagement::FaultDetector_ptr faultDetector)
#else
Host_impl::Host_impl (PortableServer::POA_ptr poa,
            const CdmwPlatformMngt::HostDef& host_def,
            const char* host_path,
            System_impl* system,
            CdmwPlatformMngt::RWElementRepository_ptr element_repository)
#endif
    throw (BadParameterException,
           OutOfMemoryException,
   	   CosPropertyService::MultipleExceptions)
    : DeactivableServant_impl (poa),
      ManagedElement_impl (element_repository, poa)
                             
{
            m_hostEntityStatusChangeFactory = NULL;
            m_actualEntityContainer = NULL;
	    m_reboot_in_progress = false;
            m_default_entitydefs = NULL;
 
            try
            { 
                if (system == NULL)
                {
                    CDMW_THROW2 (BadParameterException, "system", "NULL");
                }

                //Enable PropertyEvent features                
                set_property_event_notifier(EventNotifierManager::Instance()->get_property_notifier());
                //Enable StatusChangeEvent features
                set_status_event_notifier(EventNotifierManager::Instance()->get_status_notifier()); 

                // Set the internal data
                m_system = system; 
                m_monitor = NULL;
                // create the host entity status change factory
                std::auto_ptr <HostEntityStatusChangeFactory> hostEntityStatusChangeFactory(
                    new HostEntityStatusChangeFactory());
                
                try
                {
                    Cdmw::CommonSvcs::DataStore::DataStoreBase* ds =
                        Cdmw::CommonSvcs::DataStore::DataStoreRegistry::Get_data_store(
                        Configuration::Get_host_datastore_id());
                    m_host_ds = 
                        dynamic_cast<Cdmw::PlatformMngt::HostDataStore*>(ds);
                }
                catch (const NotExistException& e)
                {
                    CDMW_ASSERT(false);
                }
                
                
                // store element path in managed element
                set_element_path (host_path,
                    host_def.host_name.in());
                
                // create entity container               
                std::auto_ptr <PlatformMngt::EntityContainer> actualEntityContainer (
                    new PlatformMngt::EntityContainer(poa, m_element_repository.in(),
                    m_element_path.in()));
                
                m_rwEntityContainerServant = new RWEntityContainer_impl (
                    poa, actualEntityContainer.get(), m_element_repository.in());
                
                m_rwEntityContainer
                    = dynamic_cast <RWEntityContainer_impl*> (m_rwEntityContainerServant.in());
                
                // Complete the internal data
                m_hostEntityStatusChangeFactory = hostEntityStatusChangeFactory.get();
                m_actualEntityContainer = actualEntityContainer.get();
                
                // initialise data store of the object
                create_record_in_datastore(host_def);
                
                
#ifdef CDMW_USE_FAULTTOLERANCE
                m_faultDetector = CdmwFT::FaultManagement::FaultDetector::_duplicate(faultDetector);
                
                m_monitoringID = 0;        
#else
                // create Daemon PullMonitor
                std::auto_ptr <DaemonMonitor> 
                    monitor(new DaemonMonitor(   
                    m_system, 
                    m_element_name.in(),                   
                    host_def.daemon_monitoring_info.monitoring_interval,
                    host_def.daemon_monitoring_info.monitoring_timeout));
                
                
                // create Callbacks                        
                std::auto_ptr <HostRespondingCallback>  responseCallback (
                    new HostRespondingCallback (this));
                
                std::auto_ptr <HostNotRespondingCallback>  noResponseCallback (
                    new HostNotRespondingCallback (this));
                
                
                m_monitor = monitor.release();
                
                // set callbacks in monitor
                m_monitor->setResponseCallback (responseCallback.release());
                m_monitor->setNoResponseCallback (noResponseCallback.release());
#endif
                
                // Release the ownership of the temporary objects
                hostEntityStatusChangeFactory.release();
                actualEntityContainer.release();
                
                //Add the Properties on Host                                                                                                                                       
                unsigned int propCount = host_def.properties.length();
                for (unsigned int propIndex=0 ; propIndex < propCount ; propIndex++)
                {
                    const CdmwPlatformMngt::PropertyDef& prop_def =
                        host_def.properties[propIndex];
                    std::string propertyName(host_def.properties[propIndex].property_name);
                    CosPropertyService::PropertyModeType propertyMode = CosPropertyService::PropertyModeType(0);
                    CdmwPlatformMngt::PropertyModeType propertyModeCdmw = host_def.properties[propIndex].mode_type;
                    if (propertyModeCdmw == 1) propertyMode = CosPropertyService::PropertyModeType(1);
                    try
                    {
                        define_property_with_mode(propertyName.c_str() , prop_def.initial_value, propertyMode);
                    }
                    catch (...)
                    {
                        throw  CosPropertyService::MultipleExceptions(); 
                    }
                }
    }
    catch (const std::bad_alloc&)
    {
        if (m_actualEntityContainer != NULL)
            delete m_actualEntityContainer;
        
        CDMW_THROW (OutOfMemoryException);
    } 
    
}



Host_impl::~Host_impl ()
{  
    
#ifndef CDMW_USE_FAULTTOLERANCE
    if (m_monitor != NULL)
    {
        // terminate pull monitor thread if not stopped
        m_monitor->stop();
        
        delete m_monitor;
    }
#endif
    
    if (m_hostEntityStatusChangeFactory != NULL)
    {
        m_hostEntityStatusChangeFactory->_remove_ref();
    }
    
    if (m_actualEntityContainer != NULL)
    {
        delete m_actualEntityContainer;
    }

}


void Host_impl::remove()
{
    // first increment counter reference to avoid the destructor calling
    // by servant deactivation
    this->_add_ref();
    
    // deactivate the Corba object
    this->deactivate();
    
    // remove associated host record in datastore
    remove_record_in_datastore();
    
    // remove the associated managed element
    this->ManagedElement_impl::remove();
    
    // then decrement the counter reference to call destructor
    this->_remove_ref();        
}


void Host_impl::end_creation(const CdmwPlatformMngt::HostDef& host_def)
       throw (CdmwPlatformMngt::DuplicateEntity,
       OutOfMemoryException,
       CORBA::SystemException)
{
    //store the default properties
    store_default_properties();

    // Add the entities of the process
    // if duplicated throw DuplicateEntity exception
    unsigned int entityCount = host_def.entities.length();

    // Store entity initial values into default_entities
    m_default_entitydefs = new CdmwPlatformMngt::EntityDefs ();
    m_default_entitydefs->length (entityCount);
    
    for (unsigned int entityIndex=0;
    entityIndex < entityCount;
    entityIndex++)
    {  
        // get entity definition (EntityDef is a structure, a reference on the
        // structure is returned by the [] operator)
        const CdmwPlatformMngt::EntityDef& entity_def = host_def.entities[entityIndex];
        
        try
        {
            m_rwEntityContainer->add_entity(entity_def);
            // store values of the entityDef
            (*m_default_entitydefs)[entityIndex] = entity_def;

        }
        catch(const CdmwPlatformMngt::EntityAlreadyExists&)
        {
            // in this case the servant will be deleted by application
            throw CdmwPlatformMngt::DuplicateEntity (entity_def.entity_name.in());
        }
    }
}


void 
Host_impl::create_record_in_datastore(const CdmwPlatformMngt::HostDef& host_def)
    throw (OutOfMemoryException)
{
    // test if the supervision is not fault tolerant or if it is, is it the primary?
    if (!Configuration::Is_fault_tolerant() || Configuration::Is_primary())
    {
        // set managed element
        set_element_in_datastore();
        
        CdmwPlatformMngt::HostRecord record;
        record.status = CdmwPlatformMngt::HOST_RESPONDING;
        record.platform_application_exe = host_def.platform_application_exe;
        record.daemon_monitoring_info = host_def.daemon_monitoring_info;
        
        try
        {
            m_host_ds->insert(m_element_path.in(), record);
        }
        catch (Common::AlreadyExistException&)
        {
            CDMW_FATAL(Cdmw::Logging::Logger::Get_Logger(CdmwPlatformMngtBase::PFMMNG),
                       "host record already exists in datastore : " 
                       << m_element_path.in());
            CDMW_ASSERT(false);
        }  
    }
}

void 
Host_impl::remove_record_in_datastore()
{
    // inhibition if backup
    if (!Configuration::Is_fault_tolerant() || Configuration::Is_primary())
    {
        try
        {
            m_host_ds->remove(m_element_path.in());
        }
        catch (Common::NotFoundException&)
        {
        }
    }
}

CdmwPlatformMngt::HostRecord* Host_impl::get_record_from_datastore()
{
    CdmwPlatformMngt::HostRecord* host_record = NULL;
    
    try
    {
        host_record = m_host_ds->select(m_element_path.in());
    }
    catch (Common::NotFoundException&)
    {
        CDMW_FATAL(Cdmw::Logging::Logger::Get_Logger(CdmwPlatformMngtBase::PFMMNG),
                   "host record cannot be found in datastore : " 
                   << m_element_path.in());
        CDMW_ASSERT(false);
    }   
    
    return host_record; 
}



void Host_impl::update_record_in_datastore(const CdmwPlatformMngt::HostRecord& host_record)
{
    try
    {
        m_host_ds->update(m_element_path.in(), host_record);
    }
    catch (Common::NotFoundException&)
    {
        CDMW_FATAL(Cdmw::Logging::Logger::Get_Logger(CdmwPlatformMngtBase::PFMMNG),
                   "host record cannot be found in datastore : " 
                   << m_element_path.in());
        CDMW_ASSERT(false);
    } 
}



CdmwPlatformMngt::HostStatus Host_impl::status()
throw( CORBA::SystemException )
{
    CdmwPlatformMngt::HostRecord_var host_record = get_record_from_datastore();
    
    return host_record->status;
}

char * Host_impl::host_status_info() throw( CORBA::SystemException )
{
    CdmwPlatformMngt::HostRecord_var host_record = get_record_from_datastore();
    
    return ( host_record->host_status_info)._retn() ;
    
}

CdmwPlatformMngt::ManagedElement_ptr Host_impl::get_parent() throw (CORBA::SystemException)
{
    
    return m_system->_this() ;
}

    /**
     *Purpose:
     *<p> facade to get an attribute
     *
     */

     void Host_impl::get_attribute(const char * attr_name, CORBA::Any_out ret_attr, CORBA::Char flag  )  throw (CORBA::SystemException, CdmwPlatformMngt::AttributeNotFound)
        {
    char flag_found=0;
    std::string tmp1(attr_name);
    /*active the right methods*/
    
    if (flag == 0)
    { /*case of stardard attribute*/
        if ( tmp1 == "status" )
        {
            flag_found=1;
            CORBA::Any_var tempany = new CORBA::Any;
            CdmwPlatformMngt::HostStatus status_out;
            status_out = status() ; 
            tempany.inout() <<= status_out;
            ret_attr = tempany._retn();
        }
        
        if (flag_found==0)
            throw CdmwPlatformMngt::AttributeNotFound(); 
    }
    else
    { /*case of user defined attribute*/
        try
        {       
            CORBA::Any* outany = get_property_value( attr_name );
            
            CORBA::Any_var tempany = new CORBA::Any;
            
            ret_attr = outany; 
        }
        catch ( CosPropertyService::PropertyNotFound )
        {
            throw CdmwPlatformMngt::AttributeNotFound();
        }       
    }
      }

    /**
     *Purpose:
     *<p> facade to set an attribute
     *
     */

     void Host_impl::set_attribute(const char * attr_name, const CORBA::Any & attr, CORBA::Char flag  )  throw (CORBA::SystemException, CdmwPlatformMngt::AttributeNotFound)
        {
    if ( flag == 0)
    { /*no attribute to set at the moment*/
        throw CdmwPlatformMngt::AttributeNotFound();
    }
    else
    {
        /* case of user defined attributes*/
        define_property(attr_name, attr);  
    }
        }


CdmwPlatformMngt::Processes * Host_impl::get_processes ( CORBA::ULong how_many, CdmwPlatformMngt::ProcessesIterator_out processes_iterator)
throw (CORBA::SystemException)
{

  CdmwPlatformMngt::Application_var  app;

  CORBA::ULong how_many_apps; //number of applications
                         //in this system
  CORBA::ULong i;

  CdmwPlatformMngt::Applications_var  apps ;

  CdmwPlatformMngt::ApplicationsIterator_var app_iter;

  //numeric_limits <CORBA::ULong>  big_number;
  CORBA::ULong max_ulong;
  //max_ulong = big_number.max();  

  max_ulong = m_system->get_number_of_applications();
  apps = m_system->get_applications(max_ulong, app_iter) ;

  //number of applications
  how_many_apps = apps->length() ;

  CdmwPlatformMngt::Processes_var  processes;

  CORBA::ULong j,k;
  CdmwPlatformMngt::Host_ptr actual_host;
  CORBA::ULong size_seq;

  CdmwPlatformMngt::ProcessesIterator_var process_iter;

  CdmwPlatformMngt::Processes * seq_processes_out = new CdmwPlatformMngt::Processes();

  std::list<CdmwPlatformMngt::Process_var> process_remaining;

//use a cycle with the single get_application
    k=0;
    for (i=0 ; i<how_many_apps; i++)
    {
        
        max_ulong = (*apps)[i]->get_number_of_processes();
        processes = (*apps)[i]->get_processes( max_ulong, process_iter) ;
        
        size_seq = processes->length();
        
        
        for (j=0; j<size_seq; j++)
        {
            //use the get_host to know if it is us desired host
            actual_host = ((*processes)[j])->get_host();
            
            if ( actual_host->_is_equivalent(this->_this() ) )
            {
                if ( k <= (how_many-1) )
                {
                    seq_processes_out->length(k+1);
                    (*seq_processes_out)[k] =  (*processes)[j] ;
                }
                else
                {
                    /*put the processes to the iterator out ... */
                    process_remaining.push_back ( (*processes)[j] );
                    
                }
                k++;
            }
        }
        
    }
    
    CdmwPlatformMngt::ProcessesIterator_var process_remaining_iter = ProcessesIterator_impl::create( process_remaining) ;
    
    processes_iterator = process_remaining_iter._retn() ;
    
    return (seq_processes_out) ;
    
}

CdmwPlatformMngt::Proc_filters * Host_impl::get_trace_filtering()throw (CORBA::SystemException)
    {

    //see all the managed processes in the host

            CdmwPlatformMngt::Processes_var my_processes_from_host;
            CdmwPlatformMngt::ProcessesIterator_var proc_iter;

            numeric_limits < CORBA::ULong > big_number;
            CORBA::ULong max_ulong, seq_size , ii ;
            max_ulong = big_number.max ();

            my_processes_from_host =  get_processes( max_ulong, proc_iter) ;

            CdmwPlatformMngt::Proc_filters * seq_filters_out = new CdmwPlatformMngt::Proc_filters();

            //analyse all the processes ...
            seq_size = my_processes_from_host->length();


            CORBA::ULong k=0;

            for (ii=0; ii<seq_size; ii++)
               {
               //see if the process is managed ...
               if ( my_processes_from_host[ii]->type() != CdmwPlatformMngt::UNMANAGED_PROCESS )
                 {

                  //retrieve name of appl. and name of Process
                  //name of appl.
                  CdmwPlatformMngt::Application_var app = (CdmwPlatformMngt::Application:: _narrow (my_processes_from_host[ii]->get_parent ()));
		  
		  
                  //set the trace filters...
                  seq_filters_out->length(k+1);  
		  string app_name = app->name() ;
		  //string proc_name (proc_snap->process_name) ;
		  string proc_name = my_processes_from_host[ii]->name() ;
		  string out_str;
		  string filter_str;

		  if ( my_processes_from_host[ii]->get_trace_filtering() != NULL)
		    {
		    filter_str = ( my_processes_from_host[ii]->get_trace_filtering() );

		    out_str = "Application=\""+app_name+"\"; Process=\""+ proc_name + "\"; Filters=\"" + filter_str+"\"";
		    }
                  else
		    
		    out_str = "Application=\""+app_name+"\"; Process=\""+ proc_name + "\"; Filters=\"\"" ;

                 
                  (*seq_filters_out)[k] = out_str.c_str()  ;
		  
		  k++;
                 }
               }

    return ( seq_filters_out ) ;
    }


CdmwPlatformMngt::CompleteMonitoringInfo Host_impl::get_monitoring_parameters() throw (CORBA::SystemException)
  {
    CdmwPlatformMngt::HostRecord_var host_record = get_record_from_datastore();
    
    return host_record->monitoring_parameters;
  }

void Host_impl::set_monitoring_parameters(const CdmwPlatformMngt::CompleteMonitoringInfo& monitoring_parameters) throw (CORBA::SystemException)
    {

        CdmwPlatformMngt::HostRecord_var host_record;
        CdmwPlatformMngt::CompleteMonitoringInfo current_monitor_par;
    
        // lock for concurent access
        CDMW_MUTEX_GUARD(m_mutex_host_record);
    
        // get current status
        host_record = get_record_from_datastore();
        current_monitor_par = host_record->monitoring_parameters;
    
        // if current_monitor_info changes
        if ( (monitoring_parameters.monitoring_model !=
	current_monitor_par.monitoring_model) || 
	(monitoring_parameters.pull_monitoring_interval !=
	current_monitor_par.pull_monitoring_interval) ||
	(monitoring_parameters.pull_monitoring_timeout !=
	current_monitor_par.pull_monitoring_timeout) ||
	(monitoring_parameters.push_monitoring_interval !=
	current_monitor_par.push_monitoring_interval) ||
	(monitoring_parameters.push_monitoring_timeout !=
	current_monitor_par.push_monitoring_timeout) )
        {
            // store new status
            host_record->monitoring_parameters = monitoring_parameters; 
            update_record_in_datastore(host_record.in());
        }

    }

CORBA::Boolean Host_impl::get_global_supervision_indicator() throw (CORBA::SystemException)
    {
    CdmwPlatformMngt::HostRecord_var host_record = get_record_from_datastore();

    return host_record->global_supervision_indicator;

    }


 void Host_impl::set_global_supervision_indicator(bool global_supervision_indicator) throw (CORBA::SystemException)
    {
        CdmwPlatformMngt::HostRecord_var host_record;
        CORBA::Boolean current_supervision_indicator ;

        // lock for concurent access
        CDMW_MUTEX_GUARD(m_mutex_host_record);
    
        // get current status
        host_record = get_record_from_datastore();
        current_supervision_indicator = host_record->global_supervision_indicator;
    
        // if current_monitor_info changes
        if (global_supervision_indicator != current_supervision_indicator)
        {       
            // store new status
            host_record->global_supervision_indicator = global_supervision_indicator; 
            update_record_in_datastore(host_record.in());
        }       

    } 



 CdmwPlatformMngt::Process_ptr Host_impl::get_process(const char *process_name, const char *application_name ) throw (CdmwPlatformMngt::ProcessNotFound, CORBA::SystemException) 
   {
    CdmwPlatformMngt::Application_var  app;
    app = m_system->get_application(application_name);
    
    
    CdmwPlatformMngt::Host_ptr actual_host;
    
    char flag_found=0; //flag to indicate if the desired process was found
    CdmwPlatformMngt::Process_ptr myprocess;
    
    CdmwPlatformMngt::ProcessesIterator_var proc_iter;
    
    myprocess = app->get_process( process_name, "" );
    
    actual_host = myprocess->get_host() ;
    
    if ( actual_host->_is_equivalent(this->_this() ) )
    {
        flag_found=1;
        //break;
    }
    
    if (flag_found == 0)
    {
        //process not found
        throw CdmwPlatformMngt::ProcessNotFound();
    }
    
    
    //until here it means that the functions has a success  
    return CdmwPlatformMngt::Process::_duplicate(myprocess) ;
  }


void Host_impl::set_trace_filtering(const char *trace_filtering) throw (CORBA::SystemException, CdmwPlatformMngt::ParsingException , CdmwPlatformMngt::CastException ) 
	    {

	    //see all the processes in the host
	    CdmwPlatformMngt::Processes_var my_processes_from_host;
	    CdmwPlatformMngt::ProcessesIterator_var proc_iter;

	    numeric_limits < CORBA::ULong > big_number;
	    CORBA::ULong max_ulong, seq_size , ii ;
	    max_ulong = big_number.max ();

	    my_processes_from_host =  get_processes( max_ulong, proc_iter) ;

	    //analyse all the processes ...
	    seq_size = my_processes_from_host->length();


	    try
	    {

	    for (ii=0; ii<seq_size; ii++)
	       {
	       //see if the process is managed ...
	       if ( my_processes_from_host[ii]->type() != CdmwPlatformMngt::UNMANAGED_PROCESS )
		 {
		 //set the trace filters...
		  my_processes_from_host[ii]->set_trace_filtering(trace_filtering);
		 }
	       }
	    
	    }
	    catch( CdmwPlatformMngt::ParsingException& ex)
	      {
	      throw CdmwPlatformMngt::ParsingException(ex.message);
	      }
	    catch( CdmwPlatformMngt::CastException& ex)
	      {
	      throw CdmwPlatformMngt::CastException( ex.message) ;
	      }

	    }

void Host_impl::shutdown ()
throw (CORBA::SystemException)
{
   if (true == get_reboot_in_progress() ) 
   {
      return;
   }
   else 
   {
        set_reboot_in_progress(true);
   }
   CDMW_INFO(Cdmw::Logging::Logger::Get_Logger(CdmwPlatformMngtBase::PFMMNG),
             "Host Shutdown : " << m_element_path.in());

   stop_processes();

   try
   {
        CdmwPlatformMngt::Daemon_var m_daemon = Configuration::Get_daemon (m_element_name.in(),
                                      m_system->systemPort());

        m_daemon -> shutdown();
   }
   catch (...)
   {
       CDMW_ERROR(Cdmw::Logging::Logger::Get_Logger(CdmwPlatformMngtBase::PFMMNG),
            "Cannot shutdown remote host : " << m_element_path.in());
       throw;
       set_reboot_in_progress(false);

   }
   set_reboot_in_progress(false);

}



void Host_impl::cold_reboot ()
throw (CORBA::SystemException)
{
    recover_default_properties ();
    recover_default_entities (m_rwEntityContainer, m_default_entitydefs);
    {
    	if (true == get_reboot_in_progress() ) {
    	return;
    	}
    	else {
    	set_reboot_in_progress(true);
    	set_requested_restart_warmth (CdmwPlatformMngt::COLD_RESTART);
	}
    }
   CDMW_INFO(Cdmw::Logging::Logger::Get_Logger(CdmwPlatformMngtBase::PFMMNG),
             "Cold rebooting : " << m_element_path.in());

   stop_processes();

   try 
   {
        CdmwPlatformMngt::Daemon_var m_daemon = Configuration::Get_daemon (m_element_name.in(),
                                      m_system->systemPort());

     m_daemon -> cold_reboot();
   }
   catch (...)
   {
       CDMW_ERROR(Cdmw::Logging::Logger::Get_Logger(CdmwPlatformMngtBase::PFMMNG),
            "Cannot reboot remote host : " << m_element_path.in());
       throw;
       set_reboot_in_progress(false);
   }
        
}


void Host_impl::hot_reboot ()
throw (CORBA::SystemException)
{
    // TODO set to datastore
    {
    	if (true == get_reboot_in_progress() ) {
    	return;
    	}
    	else {
    	set_reboot_in_progress(true);
	set_requested_restart_warmth (CdmwPlatformMngt::HOT_RESTART); 
    	}
    }

    CDMW_INFO(Cdmw::Logging::Logger::Get_Logger(CdmwPlatformMngtBase::PFMMNG),
             "Hot rebooting : " << m_element_path.in());

    stop_processes();
 
    try {	
	CdmwPlatformMngt::Daemon_var m_daemon = Configuration::Get_daemon (m_element_name.in(),                  
                                      m_system->systemPort());

     m_daemon -> hot_reboot();
	      }
	catch (...)
	{
        CDMW_ERROR(Cdmw::Logging::Logger::Get_Logger(CdmwPlatformMngtBase::PFMMNG),
             "Cannot reboot remote host : " << m_element_path.in());
	throw;	
        set_reboot_in_progress(false);
	}
    
}


void Host_impl::stop_processes()
{
    CORBA::ULong how_many = 1L;
    CdmwPlatformMngt::ProcessesIterator_var processes_iterator;
    CdmwPlatformMngt::Processes_var processes = get_processes ( how_many, processes_iterator.out());
    CdmwPlatformMngt::Process_var first_process = processes[(CORBA::ULong)0];
    CdmwPlatformMngt::Process_var other_process;
    first_process->stop_and_wait(false);
    while ( processes_iterator -> next_one(other_process.out() ) ) 
    {
         std::cerr << " stopping all Host processes " << other_process->name()<< std::endl;
         other_process ->stop_and_wait(false);
    }
}

void Host_impl::start_processes()
{

    CORBA::ULong how_many = 1L;
    CdmwPlatformMngt::ProcessesIterator_var processes_iterator;	

    //Set default startup kind for processes to be restarted    
    CdmwPlatformMngtBase::StateRetrieval state_retrieval = CdmwPlatformMngtBase::LAST_STATE;
    unsigned long startup_mode = 0;
    string specified_state_id = "";
    CdmwPlatformMngtBase::StartupKind_var startupKind
	= new CdmwPlatformMngtBase::StartupKind();
    startupKind->state_retrieval = state_retrieval;
    startupKind->startup_mode = startup_mode;
    startupKind->specified_state_id 
	= CORBA::string_dup( specified_state_id.c_str() );

    //start processes
    CdmwPlatformMngt::Processes_var processes = get_processes ( how_many, processes_iterator.out());
    CdmwPlatformMngt::Process_var first_process = processes[(CORBA::ULong)0];
    CdmwPlatformMngt::Process_var other_process;
    
	   //Cold restart
	   if (CdmwPlatformMngt::COLD_RESTART == requested_restart_warmth() ){
                    HostProcessInitSequencer::get_instance()->initialise_host_processes(CdmwPlatformMngt::COLD_RESTART, name());
                    HostProcessInitSequencer::get_instance()->run_host_processes(CdmwPlatformMngt::COLD_RESTART,name());                
 	    	}
	    //Hot restart
            else {
                    HostProcessInitSequencer::get_instance()->initialise_host_processes(CdmwPlatformMngt::HOT_RESTART, name());
                    HostProcessInitSequencer::get_instance()->run_host_processes(CdmwPlatformMngt::HOT_RESTART, name());     
 	    	}
      
      set_requested_restart_warmth (get_default_restart_warmth());
      CDMW_INFO(Cdmw::Logging::Logger::Get_Logger(CdmwPlatformMngtBase::PFMMNG),
             "Reboot is finished of remote host : " << m_element_path.in());
}


char* Host_impl::platform_application_exe()
throw( CORBA::SystemException )
{
    CdmwPlatformMngt::HostRecord_var host_record = get_record_from_datastore();
    
    // Note : assignment to a const char* return by .in() leads to make a copy of the string.
    CORBA::String_var exe = (host_record->platform_application_exe).in();
    
    return exe._retn();
}


void Host_impl::setStatus (CdmwPlatformMngt::HostStatus status, CdmwPlatformMngtBase::TimeStamp time_stamp)
{
    CdmwPlatformMngt::HostRecord_var host_record;
    CdmwPlatformMngt::HostStatus current_status;
    
    // get current status
    host_record = get_record_from_datastore();
    current_status = host_record->status;
    
    // if status has been changed
    if (status != current_status)
    {
        // notify status change
        m_system->notifyHostStatusChangeEvent (
            get_element_name(),
            status, time_stamp);

        //Prepare the event
        // create header
        CdmwPlatformMngtBase::EventHeader header =
              EventHeaderFactory::createHeader (time_stamp,
              CdmwPlatformMngtBase::WRN);

        std::auto_ptr < HostStatusChangeFactory >
           hostStatusChangeFactory (new HostStatusChangeFactory ());

        // create event value type using its factory
        CdmwPlatformMngt::HostStatusChange_var event =
            hostStatusChangeFactory->create (header, m_element_name.in (),
                                               m_element_name.in (), status);
        // Trace to the log manager (must be done before addEvent)
        CORBA::String_var eventStr = event->to_string ();

        // Perform the status change notification to observer registered
        // with the Api in ManagedElement
        event->_add_ref();
        get_status_event_notifier()->add_event (event);
        event->_remove_ref();

        // release owner ship if no error
        event._retn ();
 

   
        if (status == CdmwPlatformMngt::HOST_UNREACHABLE)
        {
            // update applications of system with unavailable host name
            m_system->updateApplications (get_element_name());
        }
        
        // lock for concurent access
        {
        CDMW_MUTEX_GUARD(m_mutex_host_record);
        
        // store new status
        host_record->status = status;
        update_record_in_datastore(host_record.in());
        }

    }
}


void Host_impl::start()
{
    // start pull monitor thread
#ifdef CDMW_USE_FAULTTOLERANCE
    ::FT::Location location;
    location.length(1);
    location[0].kind = "hostname";
    location[0].id = get_element_name();
    
    ::FT::FaultMonitoringIntervalAndTimeoutValue time;
    time.monitoring_interval=0;
    time.timeout=0;    
    
    m_monitoringID = m_faultDetector->start_monitoring(::FT::PullMonitorable::_nil(),
        location,
        time);
#else
    m_monitor->start();
#endif
}



void Host_impl::stop()
{
    // inhibition if backup
    if (!Configuration::Is_fault_tolerant() || Configuration::Is_primary())
    {
        // terminate pull monitor thread
        
#ifdef CDMW_USE_FAULTTOLERANCE
        m_faultDetector->stop_monitoring(m_monitoringID);            
#else
        m_monitor->stop();        
#endif

    }
}


void Host_impl::notifyEntityStatusChangeEvent(
                                              const char* entity_name,
                                              const char* entity_status,
                                              const char* entity_info )
{
    try
    {
        CdmwPlatformMngtBase::EventHeader header
            = EventHeaderFactory::createHeader(CdmwPlatformMngtBase::INF);
        
        CdmwPlatformMngt::HostEntityStatusChange_var event
            = m_hostEntityStatusChangeFactory->create(
            header,
            m_system->get_element_name(),
            m_element_name.in(),
            entity_name,
            entity_status,
            entity_info);
        
        CORBA::String_var eventStr = event->to_string();
        LogMngr::logMessage(eventStr.in());
     
        event->_add_ref();   
        m_system->get_event_notifier()->addEvent(event);

        // Perform the status change notification to observer registered
        // with the Api in ManagedElement
        event->_add_ref();
        get_status_event_notifier()->add_event (event);
        event->_remove_ref();
  
        event._retn();
    }
    catch(const std::bad_alloc&)
    {
        throw CORBA::NO_MEMORY(OrbSupport::NO_MEMORY, CORBA::COMPLETED_NO);
    }
    catch(...)
    {
        throw CORBA::INTERNAL(OrbSupport::INTERNAL, CORBA::COMPLETED_NO);
    }
}

CORBA::ULong Host_impl::get_number_of_entities()
throw(CORBA::SystemException)
{
    return m_rwEntityContainer->get_number_of_entities();
}

void Host_impl::get_all_entity_names(
                                     CORBA::ULong how_many,
                                     CdmwPlatformMngt::ElementNames_out entity_names,
                                     CdmwPlatformMngt::ElementNamesIterator_out rest)
throw(CORBA::SystemException)
{
    m_rwEntityContainer->get_all_entity_names(
        how_many, entity_names, rest);
}

CdmwPlatformMngt::Entity_ptr Host_impl::get_entity (
                                                    const char* entity_name)
throw (CdmwPlatformMngt::EntityNotFound,
                                                    CORBA::SystemException)
{
    return m_rwEntityContainer->get_entity(entity_name);
}

CORBA::Boolean Host_impl::get_entities(
                                       const CdmwPlatformMngt::ElementNames& entity_names,
                                       CdmwPlatformMngt::Entities_out nentities)
throw(CORBA::SystemException)
{
    return m_rwEntityContainer->get_entities(
        entity_names, nentities);
}

void Host_impl::get_all_entities(
                                 CORBA::ULong how_many,
                                 CdmwPlatformMngt::Entities_out nentities,
                                 CdmwPlatformMngt::EntitiesIterator_out rest)
throw(CORBA::SystemException)
{
    m_rwEntityContainer->get_all_entities(
        how_many, nentities, rest);
}


CdmwPlatformMngt::Entity_ptr Host_impl::add_entity (
                                                    const CdmwPlatformMngt::EntityDef & entity_def)
throw(CdmwPlatformMngt::EntityAlreadyExists,
                                                    CORBA::SystemException)
{
    std::string entityName = entity_def.entity_name.in();
    
    LogMngr::logMessage(INF, MSG_ID_ADD_HOST_ENTITY, 
        entityName.c_str(), m_element_name.in());
    
    CdmwPlatformMngt::Entity_ptr ent_ret = m_rwEntityContainer->add_entity(entity_def);
    
    notify_configuration_change("Entity Added");
    return ent_ret; 
}

void Host_impl::remove_entity(
                              const char* entity_name)
throw (CdmwPlatformMngt::EntityNotFound,
                              CORBA::SystemException)
{
    LogMngr::logMessage(INF, MSG_ID_REMOVE_SYSTEM_ENTITY, 
        entity_name, m_element_name.in());
    
    m_rwEntityContainer->remove_entity(entity_name);
    
    notify_configuration_change("Entity Removed");
}
CdmwPlatformMngt::ElementType Host_impl::get_me_type()
    throw (CORBA::SystemException)
{
    return CdmwPlatformMngt::HOST_TYPE;
}

CdmwPlatformMngt::Children* Host_impl::get_children()
    throw (CORBA::SystemException)
{
     CdmwPlatformMngt::Children_var children = new CdmwPlatformMngt::Children;
     CORBA::ULong size_of_entities = get_number_of_entities();
     children->length(size_of_entities);

     CdmwPlatformMngt::Entities_var entities;
     CdmwPlatformMngt::EntitiesIterator_var entities_iterator;
     get_all_entities(size_of_entities, entities.out(), entities_iterator.out() );

     for (CORBA::ULong count = 0; count<size_of_entities ; count++)
     {
         entities[count] -> _add_ref();
         children[count] = (entities[count])._retn();
     }
     return children._retn ();
}

void  Host_impl::register_observer_ext(const char* observer_name,
                                       CdmwPlatformMngt::RegistrationDepth depth,
                                       CdmwPlatformMngt::SupervisionObserver_ptr observer)  throw (CORBA::SystemException,
                                                                                                   CdmwPlatformMngt::ElementAlreadyExists)
        {
             ManagedElement_impl::register_observer_ext(observer_name, depth, observer);
             if (depth==CdmwPlatformMngt::FULL_HIERARCHY)
             {
                 CdmwPlatformMngt::ProcessesIterator_var proc_iter;
                 numeric_limits < CORBA::ULong > big_number;
                 CORBA::ULong max_ulong;
                 max_ulong = big_number.max ();
                 CdmwPlatformMngt::Processes_var processes_from_host = get_processes (max_ulong, proc_iter.out());
                 CORBA::ULong size1 = processes_from_host->length();
                 CORBA::ULong size = 0;
                 while (size!=size1)
                 {  
                     size++;
                     processes_from_host[size-1]->register_observer_ext(observer_name, depth, CdmwPlatformMngt::SupervisionObserver::_duplicate(observer));
                 }
             }
        }

void Host_impl::unregister_observer_ext(const char* observer_name) throw (CORBA::SystemException, 
                                                                          CdmwPlatformMngt::ObserverNotFound)
{
    ManagedElement_impl::unregister_observer_ext(observer_name);
    CdmwPlatformMngt::ProcessesIterator_var proc_iter;
    numeric_limits < CORBA::ULong > big_number;
    CORBA::ULong max_ulong;
    max_ulong = big_number.max ();
    CdmwPlatformMngt::Processes_var processes_from_host = get_processes (max_ulong, proc_iter.out());
    CORBA::ULong size1 = processes_from_host->length();
    CORBA::ULong size = 0;
    while (size!=size1)
    {
        size++;
        try 
        {
            processes_from_host[size-1]->unregister_observer_ext(observer_name);
        }      
        catch (const CdmwPlatformMngt::ObserverNotFound& ex)
        {
        }
    }
}

CORBA::Boolean Host_impl::get_reboot_in_progress() throw (CORBA::SystemException)
{
    CDMW_MUTEX_GUARD(m_mutexRebootInProgress);
    return m_reboot_in_progress;
}



void  Host_impl::set_reboot_in_progress(CORBA::Boolean reboot_in_progress) throw (CORBA::SystemException)
{
    CDMW_MUTEX_GUARD(m_mutexRebootInProgress);
    m_reboot_in_progress = reboot_in_progress;
}


void Host_impl::set_run_entry(const std::string& host_name, ProcessHostInfo proc_info)
{
    HostProcessInitSequencer::get_instance()->set_run_entry(host_name, proc_info);
}

void Host_impl::remove_run_entry(const std::string& host_name, ProcessHostInfo proc_info)
{
     HostProcessInitSequencer::get_instance()->remove_run_entry(host_name, proc_info);
}

void Host_impl::set_init_entry(const std::string& host_name, ProcessHostInfo proc_info)
{
     HostProcessInitSequencer::get_instance()->set_init_entry(host_name, proc_info);
}

void Host_impl::remove_init_entry(const std::string& host_name, ProcessHostInfo proc_info)
{
     HostProcessInitSequencer::get_instance()->remove_init_entry(host_name, proc_info);
}



void Host_impl::synchro_with_datastore ()
{
     Cdmw::PlatformMngt::ManagedElement_impl::synchro_with_datastore ();
}


} // End namespace PlatformMngt
} // End namespace Cdmw


