/* ========================================================================== *
 * This file is part of CARDAMOM (R) which is jointly developed by THALES
 * and SELEX Sistemi Integrati.
 * Copyright (C) SELEX Sistemi Integrati 2004-2006. All rights reserved.
 *  
 * CARDAMOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * CARDAMOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public
 * License for more details.
 *  
 *  
 * You should have received a copy of the GNU Library General Public
 * License along with CARDAMOM; see the file COPYING. If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * ========================================================================= */

#include "clockservice/Controller_impl.hpp"
#include "clockservice/Executor_impl.hpp"
#include "clockservice/ControllerUpdateAdmin_impl.hpp"
#include "clockservice/ClockTypes.hpp"

#include "Time/clockservice/Util.hpp"
#include "Time/clockservice/Time.hpp"

#include <Foundation/common/Assert.hpp>

#include <iostream>
#include <cmath>

using namespace Cdmw::clock;




///////////////////////////////////////////////////////////////////////////////
//                 -- Timeout Handler _implementation --
///////////////////////////////////////////////////////////////////////////////

Controller_impl::TimeoutHandler::TimeoutHandler(Controller_impl*                                 controller,
                                                CosClockService::PeriodicExecution::Periodic_ptr periodic,
                                                ACE_Recursive_Thread_Mutex&                      mutex)
    : controller_(controller),
      periodic_(CosClockService::PeriodicExecution::Periodic::_duplicate(periodic)),
      executions_(-1),      // NB: not meaningful here
      executionsLimit_(-1), // NB: not meaningful here
      controller_mutex_(mutex),
      timeout_handler_mutex_()
{
    // No-Op
}

///////////////////////////////////////////////////////////////////////////////

Controller_impl::TimeoutHandler::~TimeoutHandler() 
{
    // No-Op
}

///////////////////////////////////////////////////////////////////////////////

int 
Controller_impl::TimeoutHandler::handle_timeout(const ACE_Time_Value& /*current_time*/, const void* /* act */)
{
    // Handle timeout for the current execution (call to periodic callback).
    try
    {
        // Check for overflow conditions on the executions number.
        ACE_Guard<ACE_Recursive_Thread_Mutex> guard(timeout_handler_mutex_);
        long executions_done = (long) executions_;
        bool valid_execution = (executionsLimit_ == INF_EXEC_LIMIT || executions_done < executionsLimit_);
        bool last_execution  = (executionsLimit_ != INF_EXEC_LIMIT && (executions_done + 1) == executionsLimit_);

        if (valid_execution)
        {
            // Invoke Periodic::do_work()
            CORBA::Boolean do_work_result    = false;
            bool           do_work_exception = false;
            try
            {
                do_work_result = periodic_->do_work(controller_->m_any);
            }
            catch (const CORBA::SystemException& e)
            {
                // No action required in case of exception CORBA::SystemException.
                std::cerr << "Controller_impl::TimeoutHandler::handle_timeout: "
                          << "CORBA::SystemException exception raised while delivering timeout: "
                          << e << std::endl;
                do_work_exception = true;
            }
            catch (const std::exception& e)
            {
                // No action required in case of exception std::exception.
                std::cerr << "Controller_impl::TimeoutHandler::handle_timeout: "
                          << "std exception raised: "
                          << e.what() << std::endl;
                do_work_exception = true;
            }
            catch (...)
            {
                // No action required in case of generic exception.
                std::cerr << "Controller_impl::TimeoutHandler::handle_timeout: "
                          << "generic exception raised" << std::endl;                
                do_work_exception = true;
            }

            if (do_work_exception)
            {
                controller_->cancelTimer();
            }
            else
            {
                // Update the periodic executions count.
                this->executions_++;

                {
                    // Notify the controller internal status, (only for FTControllers).
                    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(controller_mutex_);
                    controller_->notify_status_lock_free();
                }

                // In case of last execution, cancel timer.
                if (last_execution || !do_work_result)
                    controller_->cancelTimer();
            }
        }
    }
    catch (...)
    {
        // No action required in case of generic exception.
        std::cerr << "Controller_impl::TimeoutHandler::handle_timeout: generic exception raised" << std::endl;
    }

    // Return to the caller.
    return 0;
}

///////////////////////////////////////////////////////////////////////////////

long& 
Controller_impl::TimeoutHandler::executionLimit()
{
    return executionsLimit_;   
}

///////////////////////////////////////////////////////////////////////////////

const long& 
Controller_impl::TimeoutHandler::executionLimit() const
{
    return executionsLimit_;   
}

///////////////////////////////////////////////////////////////////////////////

TimerAtomicCount&
Controller_impl::TimeoutHandler::executions()
{
    return executions_;   
}




///////////////////////////////////////////////////////////////////////////
//                 -- Controller_impl _implementation --

Controller_impl::Controller_impl(Executor_impl&                                   clock,
                                 CosClockService::PeriodicExecution::Periodic_ptr periodic,
                                 ACE_Reactor*                                     reactor)
    : m_mutex(),
      m_ID(Cdmw::clock::policy::UNDEFINED),
      m_clock(clock),
      m_reactor(reactor),
      m_handler(this, periodic, m_mutex),
      m_timer_key(-1),
      m_status(Cdmw::clock::CREATED),
      m_controller_time_origin(0),  // NB: not meaningful here
      m_real_period(0),             // NB: not meaningful here
      m_any()                       // NB: not meaningful here
{
    // Notify a change in the controller internal state, (only for FTControllers)
    this->notify_status_lock_free();
}

///////////////////////////////////////////////////////////////////////////////

Controller_impl::~Controller_impl()
{
    this->cancelTimer();
}

///////////////////////////////////////////////////////////////////////////////

void
Controller_impl::start(TimeBase::TimeT   period, 
                       TimeBase::TimeT   offset,
                       CORBA::ULong      execution_limit,
                       const CORBA::Any& params)
    throw (CORBA::SystemException)
{
    // If the timer had already been programmed, this is 
    // interpreted as a re-programmation. Thus, the former 
    // alarm is reset to this.
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    if ((m_status != Cdmw::clock::CREATED) && (m_status != Cdmw::clock::STARTED))
    {
        std::cerr << "ERROR: Controller is not in the CREATED or STARTED status" << std::endl;
        throw CORBA::BAD_INV_ORDER();
    }

    this->cancelTimerLockFree();

    m_status                 = Cdmw::clock::STARTED;
    m_controller_time_origin = m_clock.current_time() + offset;
    m_real_period            = period; 
    m_any                    = params;

    m_handler.executionLimit() = (long) execution_limit;
    m_handler.executions()     = TimerAtomicCount(0);

    this->scheduleTimerLockFree(period, offset);

    // Notify a change in the controller internal state, (only for FTControllers)
    this->notify_status_lock_free(); 
}

///////////////////////////////////////////////////////////////////////////////

void 
Controller_impl::start_at(TimeBase::TimeT   period ,
                          TimeBase::TimeT   at_time,
                          CORBA::ULong      execution_limit,
                          const CORBA::Any& params)

    throw (CORBA::SystemException,
           CosClockService::PeriodicExecution::Controller::TimePast)
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    TimeBase::TimeT checkedOffset = checked_offset(at_time);
    TimeBase::TimeT now = m_clock.current_time();
    
    // If checkedOffset is >> 1 then probably can assert 
    CDMW_ASSERT(now + checkedOffset >= now);

    this->start(period, checkedOffset, execution_limit, params);

    // Notify a change in the controller internal state, (only for FTControllers)
    this->notify_status_lock_free(); 
}

///////////////////////////////////////////////////////////////////////////////

void 
Controller_impl::pause()
    throw (CORBA::SystemException)
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    if (m_status != Cdmw::clock::STARTED)
    {
        std::cerr << "ERROR: Controller is not in the STARTED status" << std::endl; 
        throw CORBA::BAD_INV_ORDER();
    }

    this->cancelTimerLockFree();
    m_status = Cdmw::clock::PAUSED;

    // Notify a change in the controller internal state, (only for FTControllers)
    this->notify_status_lock_free(); 
}

///////////////////////////////////////////////////////////////////////////////

void 
Controller_impl::resume()
    throw (CORBA::SystemException)
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    if (m_status != Cdmw::clock::PAUSED)
    {
        std::cerr << "ERROR: Controller is not in the PAUSED status" << std::endl; 
        throw CORBA::BAD_INV_ORDER();
    }

    this->scheduleTimerLockFree(m_real_period, compute_next_trigger_offset());
    m_status = Cdmw::clock::STARTED;

    // Notify a change in the controller internal state, (only for FTControllers)
    this->notify_status_lock_free(); 
}

///////////////////////////////////////////////////////////////////////////////

void 
Controller_impl::resume_at(TimeBase::TimeT at_time)
    throw (CORBA::SystemException,
           CosClockService::PeriodicExecution::Controller::TimePast)
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);
    
    if (m_status != Cdmw::clock::PAUSED)
    {
        std::cerr << "ERROR: Controller is not in the PAUSED status" << std::endl; 
        throw CORBA::BAD_INV_ORDER();
    }

    this->scheduleTimerLockFree(m_real_period, checked_offset(at_time));
    m_status = Cdmw::clock::STARTED;

    // Notify a change in the controller internal state, (only for FTControllers)
    this->notify_status_lock_free(); 
}

///////////////////////////////////////////////////////////////////////////////

void 
Controller_impl::terminate()
    throw (CORBA::SystemException)
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    m_status = Cdmw::clock::TERMINATED;
    this->cancelTimerLockFree();
    m_clock.detach_terminated_controller(this);

    // Notify a change in the controller internal state, (only for FTControllers)
    this->notify_status_lock_free(); 
}

///////////////////////////////////////////////////////////////////////////////

Cdmw::clock::policy::ControllerUpdateID 
Controller_impl::get_update_policy()
    throw (CORBA::SystemException)
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    return this->m_ID;
}

////////////////////////////////////////////////////////////////////////////////

TimeBase::TimeT
Controller_impl::get_controller_time_origin()
    throw (CORBA::SystemException)
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    return m_controller_time_origin;
}      

///////////////////////////////////////////////////////////////////////////////

CORBA::ULong
Controller_impl::executions()
    throw (CORBA::SystemException)
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    return (CORBA::ULong) ((long) m_handler.executions());
}

////////////////////////////////////////////////////////////////////////////////

void
Controller_impl::set_update_policy(::Cdmw::clock::policy::ControllerUpdateID id)
    throw (CORBA::SystemException,
           ::Cdmw::clock::policy::ControllerUpdateAdmin::Unavailable)
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    this->m_ID = id;
    Cdmw::clock::policy::ControllerUpdateAdmin_impl::instance()->get_policy(this->m_ID); 
}

////////////////////////////////////////////////////////////////////////////////

void
Controller_impl::set_controller_time_origin(TimeBase::TimeT time_origin) 
    throw (CORBA::SystemException)
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    m_controller_time_origin = time_origin;
}

////////////////////////////////////////////////////////////////////////////////

Cdmw::clock::Executor_impl*
Controller_impl::retrieve_executor()
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    return &m_clock;
}

////////////////////////////////////////////////////////////////////////////////C

Cdmw::clock::Status
Controller_impl::get_status()
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    return m_status; 
}

////////////////////////////////////////////////////////////////////////////////

TimeBase::TimeT 
Controller_impl::get_period()
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    return  m_real_period;
}

////////////////////////////////////////////////////////////////////////////////

CORBA::Any
Controller_impl::get_parameter()
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    return m_any;  
}

////////////////////////////////////////////////////////////////////////////////

CORBA::ULong
Controller_impl::execution_limit()
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    return (CORBA::ULong) (m_handler.executionLimit());
}

////////////////////////////////////////////////////////////////////////////////

CosClockService::PeriodicExecution::Periodic_ptr
Controller_impl::periodic()
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    return m_handler.get_periodic();
}

////////////////////////////////////////////////////////////////////////////////

TimeBase::TimeT
Controller_impl::compute_next_trigger_offset()
{
    double rate = checked_rate();

    TimeBase::TimeT vt = m_clock.current_time();
    TimeBase::TimeT vx, v1_vx, vx_v0;
    if (vt > m_controller_time_origin)
    {
        vx    = vt - m_controller_time_origin;
        vx_v0 = vx % m_real_period;
        v1_vx = m_real_period - vx_v0;
    }
    else
    {
        vx    = m_controller_time_origin - vt;
        v1_vx = vx % m_real_period;
        vx_v0 = m_real_period - v1_vx;
    }
    
    if (rate > 0) return v1_vx;
    if (rate < 0) return vx_v0;
    return 0; // NB: just to make the compiler happy...
}

////////////////////////////////////////////////////////////////////////////////

double
Controller_impl::checked_rate()
    throw (CORBA::BAD_PARAM)
{
    double rate = m_clock.retrieve_rate();
    if (rate == 0.0) throw CORBA::BAD_PARAM();
    return rate;
}

////////////////////////////////////////////////////////////////////////////////

TimeBase::TimeT
Controller_impl::checked_offset(TimeBase::TimeT at_time)
{
    TimeBase::TimeT now    = m_clock.current_time();
    TimeBase::TimeT offset = 0;

    // Controlled clock can have a negative rate
    if (checked_rate() > 0.0)
    {
        if (now < at_time)
            offset = at_time - now;
        else
        {
            //std::cerr << "Controller_impl::checked_offset: positive rate: (now - at_time) (sec): "
            //          << Cdmw::clock::util::TimeT2seconds(now - at_time) << std::endl;
            throw CosClockService::PeriodicExecution::Controller::TimePast();
        }
    }
    else // if here, rate < 0.0
    {
        if (at_time < now)
            offset = now - at_time;
        else
        {
            //std::cerr << "Controller_impl::checked_offset: negative rate: (at_time - now) (sec): "
            //          << Cdmw::clock::util::TimeT2seconds(at_time - now) << std::endl;
            throw CosClockService::PeriodicExecution::Controller::TimePast();
        }
    }
    return offset;
}

////////////////////////////////////////////////////////////////////////////////

void    
Controller_impl::update_period(TimeBase::TimeT period)
{
    double        rate         = fabs(checked_rate());
    double        tmp_period   = period / (10.0 * rate);
    unsigned long ulong_period = (unsigned long) tmp_period;

    // Workaround for ACE problem
    if (period < 1 || tmp_period < 1.0)
        ulong_period = 1;

    m_period.set(0, ulong_period);
}

////////////////////////////////////////////////////////////////////////////////

void    
Controller_impl::update_offset(TimeBase::TimeT offset)
{
    ldouble rate = fabs(checked_rate());
    // digits of ldouble are the same that ulonglong (TimeT)
    ldouble tmp_offset = offset;
    // from this point we are in microseconds
    tmp_offset/=(10.0*rate);

    // we want to have integer value so fractional
    // values are meaningless
    if( tmp_offset < 0 )
        tmp_offset = 0;

    // like before, max value for ulonglong is represented
    // from digits
    if( tmp_offset >= Cdmw::clock::macro::MAX_TIMET )
        tmp_offset = Cdmw::clock::macro::MAX_TIMET;
   
    const TimeBase::TimeT base = 1000000; 
    TimeBase::TimeT converted = (TimeBase::TimeT)tmp_offset; 
    
    TimeBase::TimeT tt_microsecs = converted % base;
    TimeBase::TimeT tt_secs = (converted - tt_microsecs)/base;

    CDMW_ASSERT( tt_microsecs <= base );
    long int microsecs = (long int)tt_microsecs;

    CDMW_ASSERT( (TimeBase::TimeT)(microsecs) == tt_microsecs );
    const TimeBase::TimeT MAX_LONG_INT = std::numeric_limits<long int>::max();
    
    // if tt_secs exceed value max value is changed
    if( tt_secs > MAX_LONG_INT)
        tt_secs = MAX_LONG_INT;
    
    // time_t and suseconds_t are long int 
    long int secs = (long int)tt_secs;

    CDMW_ASSERT( secs >= 0 );

    m_delay.set(secs,microsecs );
}

///////////////////////////////////////////////////////////////////////////////

void  
Controller_impl::scheduleTimerLockFree(TimeBase::TimeT period, TimeBase::TimeT offset)
{
    if (m_timer_key == -1)
    {
        this->update_offset(offset);
        this->update_period(period);
        m_timer_key = m_reactor->schedule_timer(&m_handler, 0, m_delay, m_period);
        CDMW_ASSERT(m_timer_key != -1);
    }
}

///////////////////////////////////////////////////////////////////////////////

void
Controller_impl::cancelTimer() 
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    this->cancelTimerLockFree();
}

///////////////////////////////////////////////////////////////////////////////

void
Controller_impl::cancelTimerLockFree()
{
    // No need to lock since this method is always called by method
    // that have already aquired the mutex.

    // This check can occur because if an exception is thrown m_timer_key is -1
    if (m_timer_key != -1)
    {
        try
        {
            if ((m_reactor->cancel_timer(m_timer_key)) != 1)
                std::cerr << "Controller_impl::cancelTimerLockFree: ERROR: timer was not canceled properly" << std::endl;
        }
        catch (...)
        {
            std::cerr << "Controller_impl::cancelTimerLockFree: ERROR: unexpected generic exception during ACE_Reactor::cancel_timer(.)" << std::endl;
        }
        m_timer_key = -1;
    }
}

///////////////////////////////////////////////////////////////////////////////
    
void
Controller_impl::freezeTimer() 
{
    ACE_Guard<ACE_Recursive_Thread_Mutex> guard(this->m_mutex);

    this->freezeTimerLockFree();
}

///////////////////////////////////////////////////////////////////////////////////

void
Controller_impl::freezeTimerLockFree()
{
    // No need to lock since this method is always called by method
    // that have already aquired the mutex.

    // This check can occur because if an exception is thrown m_timer_key is -1
    if (m_timer_key != -1) 
    {
        try
        {
            if ((m_reactor->cancel_timer(m_timer_key)) != 1)
                std::cerr << "Controller_impl::freezeTimerLockFree: ERROR: timer was not canceled properly" << std::endl;
        }
        catch (...)
        {
            std::cerr << "Controller_impl::freezeTimerLockFree: ERROR: unexpected generic exception during ACE_Reactor::cancel_timer(.)" << std::endl;
        }
        m_timer_key = -1;
    }
}

////////////////////////////////////////////////////////////////////////////////

void
Controller_impl::notify_status_lock_free()
{
    // No-Op
}

////////////////////////////////////////////////////////////////////////////////

void
Controller_impl::refresh_data_lock_free(Cdmw::clock::ControllerData& data)
{
    // Prepare the controller data structure
    data.timer_info.controller_status           = m_status;
    data.timer_info.controller_time_origin      = m_controller_time_origin;
    data.timer_info.controller_period           = m_real_period;
    data.timer_info.controller_executions_limit = m_handler.executionLimit();
    data.timer_info.controller_executions       = (CORBA::ULong) m_handler.executions();
    data.timer_info.controller_parameter        = m_any;
    data.periodic                               =
    CosClockService::PeriodicExecution::Periodic::_duplicate(m_handler.get_periodic());
}

////////////////////////////////////////////////////////////////////////////////

void
Controller_impl::restore_object_lock_free(const Cdmw::clock::ControllerData& data)
{
    switch (data.timer_info.controller_status)
    {
        case Cdmw::clock::CREATED:
                                    break;
        case Cdmw::clock::STARTED:
                                    this->m_status                   = data.timer_info.controller_status;
                                    this->m_controller_time_origin   = data.timer_info.controller_time_origin;
                                    this->m_real_period              = data.timer_info.controller_period; 
                                    this->m_handler.executionLimit() = data.timer_info.controller_executions_limit;
                                    this->m_handler.executions()     = TimerAtomicCount(data.timer_info.controller_executions);
                                    this->m_any                      = data.timer_info.controller_parameter;
                                    this->m_handler.set_periodic(data.periodic.in());
                                    this->scheduleTimerLockFree(m_real_period, compute_next_trigger_offset());
                                    break;
        case Cdmw::clock::PAUSED:
                                    this->m_status                   = data.timer_info.controller_status;
                                    this->m_controller_time_origin   = data.timer_info.controller_time_origin;
                                    this->m_real_period              = data.timer_info.controller_period; 
                                    this->m_handler.executionLimit() = data.timer_info.controller_executions_limit;
                                    this->m_handler.executions()     = TimerAtomicCount(data.timer_info.controller_executions);
                                    this->m_any                      = data.timer_info.controller_parameter;
                                    this->m_handler.set_periodic(data.periodic.in());
                                    this->cancelTimerLockFree();
                                    break;
        case Cdmw::clock::TERMINATED:
                                    this->m_status = data.timer_info.controller_status;
                                    this->cancelTimerLockFree();
                                    this->m_clock.detach_terminated_controller(this);
                                    break;
    }
}
