/* ========================================================================== *
 * This file is part of CARDAMOM (R) which is jointly developed by THALES
 * and SELEX Sistemi Integrati.
 * Copyright (C) SELEX Sistemi Integrati 2004-2006. All rights reserved.
 *  
 * CARDAMOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * CARDAMOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public
 * License for more details.
 *  
 *  
 * You should have received a copy of the GNU Library General Public
 * License along with CARDAMOM; see the file COPYING. If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * ========================================================================= */

#include <Foundation/ossupport/OS.hpp>
#include <Foundation/orbsupport/CORBA.hpp>
#include <Foundation/orbsupport/OrbSupport.hpp>
#include <ConfAndPlug/cdmwinit/CdmwInterface.hpp>
#include <FaultTolerance/ftinit/FTServiceInit.hpp>

#include "ftclockcommon/GroupsCreatorProcessBehaviour.hpp"
#include "clockservice/StateTransferConfig.hpp"

int main(int argc, char* argv[])
{
    bool point_to_point = true;
    CORBA::ORB_var orb;

    try
    {
        // Initialise FT service
        Cdmw::CdmwInit::CDMW_preORBInit(argc, argv);

        // Set ORB strategies
        Cdmw::OrbSupport::StrategyList strategyList;
        strategyList.add_OrbThreaded();
        strategyList.add_PoaThreadPerConnection();

        // Initialises the ORB
        orb = Cdmw::OrbSupport::OrbSupport::ORB_init(argc, argv, strategyList);

        // Get the root POA
        CORBA::Object_var obj = orb->resolve_initial_references("RootPOA");
        PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(obj.in());

        PortableServer::POAManager_var poaMgr = rootPOA->the_POAManager();
        poaMgr->activate();

        // Get hosts names
        std::string host_name1 = Cdmw::OsSupport::OS::get_option_value(argc, argv, "--host1");
        std::string host_name2 = Cdmw::OsSupport::OS::get_option_value(argc, argv, "--host2");
        std::string host_name3 = Cdmw::OsSupport::OS::get_option_value(argc, argv, "--host3");

        std::cerr << "host1 = " << host_name1 << std::endl;
        std::cerr << "host2 = " << host_name2 << std::endl;
        std::cerr << "host3 = " << host_name3 << std::endl;

        if ((host_name1 == "no") || (host_name2 == "no") || (host_name3 == "no"))
        {
            std::cerr << "   -------- GroupsCreator: ERROR: not all host names are filled" << std::endl;
            std::cerr << "   -------- GroupsCreator: please check xml profile" << std::endl;
            return EXIT_FAILURE;
        }

        // Get executor processes names
        std::string executor_proc_name1 = Cdmw::OsSupport::OS::get_option_value(argc, argv, "--executorproc1");
        std::string executor_proc_name2 = Cdmw::OsSupport::OS::get_option_value(argc, argv, "--executorproc2");
        std::string executor_proc_name3 = Cdmw::OsSupport::OS::get_option_value(argc, argv, "--executorproc3");

        std::cerr << "executor1 = " << executor_proc_name1 << std::endl;
        std::cerr << "executor2 = " << executor_proc_name2 << std::endl;
        std::cerr << "executor3 = " << executor_proc_name3 << std::endl;

        if ((executor_proc_name1 == "no") || (executor_proc_name2 == "no") || (executor_proc_name3 == "no"))
        {
            std::cerr << "   -------- GroupsCreator: ERROR: not all executor process names are filled" << std::endl;
            std::cerr << "   -------- GroupsCreator: please check xml profile" << std::endl;
            return EXIT_FAILURE;
        }

        bool periodic_processes_specified = true;
        std::string periodic_proc_name1 = Cdmw::OsSupport::OS::get_option_value(argc, argv, "--periodicproc1");
        std::string periodic_proc_name2 = Cdmw::OsSupport::OS::get_option_value(argc, argv, "--periodicproc2");
        std::string periodic_proc_name3 = Cdmw::OsSupport::OS::get_option_value(argc, argv, "--periodicproc3");

        std::cerr << "periodic1 = " << periodic_proc_name1 << std::endl;
        std::cerr << "periodic2 = \"" << periodic_proc_name2 << "\"" << std::endl;
        std::cerr << "periodic3 = " << periodic_proc_name3 << std::endl;

        if ((periodic_proc_name1 == "no") && (periodic_proc_name2 == "no") && (periodic_proc_name3 == "no"))
            periodic_processes_specified = false;
        else if ((periodic_proc_name1 == "no") || (periodic_proc_name2 == "no") || (periodic_proc_name3 == "no"))
        {
            std::cerr << "   -------- GroupsCreator: ERROR: not all periodic process names are filled" << std::endl;
            std::cerr << "   -------- GroupsCreator: please check xml profile" << std::endl;
            return EXIT_FAILURE;
        }

        std::cerr << "periodic_processes_specified" << periodic_processes_specified << std::endl;

        // Creates the process behaviour
        typedef Cdmw::clock::svcs::GroupsCreatorProcessBehaviour PRCBehaviour;
        std::auto_ptr<PRCBehaviour> process_behaviour_auto_ptr(new PRCBehaviour(orb.in(),
                                                                                point_to_point,
                                                                                host_name1,
                                                                                host_name2,
                                                                                host_name3));

        // Link the executor processes to the hosts
        process_behaviour_auto_ptr->link_executor_process_and_host(executor_proc_name1, host_name1);
        process_behaviour_auto_ptr->link_executor_process_and_host(executor_proc_name2, host_name2);
        process_behaviour_auto_ptr->link_executor_process_and_host(executor_proc_name3, host_name3);

        // Link the periodic processes to the hosts
        if (periodic_processes_specified)
        {
            process_behaviour_auto_ptr->link_periodic_process_and_host(periodic_proc_name1, host_name1);
            process_behaviour_auto_ptr->link_periodic_process_and_host(periodic_proc_name2, host_name2);
            process_behaviour_auto_ptr->link_periodic_process_and_host(periodic_proc_name3, host_name3);
        }

        // Initialize Cardamom
        Cdmw::CdmwInit::CDMW_init(orb.in(), argc, argv, process_behaviour_auto_ptr.get());

        // Enter in the main loop
        orb->run();
    }
    catch (const CORBA::Exception &e)
    {
        std::cerr << __FILE__ << " " << __LINE__ << ": CORBA exception raised: " << e._name() << std::endl;
        return EXIT_FAILURE;
    }
    catch (const Cdmw::Exception &e)
    {
        std::cerr << __FILE__ << " " << __LINE__ << ": Cdmw exception raised: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch ( ... )
    {
        std::cerr << __FILE__ << " " << __LINE__ << ": unexpected exception" << std::endl;
        return EXIT_FAILURE;
    }

    if (!CORBA::is_nil(orb.in()))
    {
        try
        {
            Cdmw::CdmwInit::CDMW_cleanup(orb.in());
            Cdmw::OrbSupport::OrbSupport::ORB_cleanup(orb.in());
            orb->destroy();
        }
        catch (const CORBA::Exception& e)
        {
            std::cerr << __FILE__ << " " << __LINE__ << ": unexpected CORBA exception: " << e._name() << std::endl;
            return EXIT_FAILURE;
        }
    }

    // Exit the main program
    return EXIT_SUCCESS;
}
