/* ========================================================================== *
 * This file is part of CARDAMOM (R) which is jointly developed by THALES
 * and SELEX Sistemi Integrati.
 * Copyright (C) SELEX Sistemi Integrati 2004-2006. All rights reserved.
 *  
 * CARDAMOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * CARDAMOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public
 * License for more details.
 *  
 *  
 * You should have received a copy of the GNU Library General Public
 * License along with CARDAMOM; see the file COPYING. If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * ========================================================================= */

#include "ftclockcommon/GroupsCreatorProcessBehaviour.hpp"
#include "clockservice/StateTransferConfig.hpp"

#include <FaultTolerance/idllib/CdmwFTReplicationManager.stub.hpp>
#include <FaultTolerance/idllib/FT.stub.hpp>
#include <FaultTolerance/ftcommon/FactoryInfos.hpp>
#include <FaultTolerance/ftcommon/FTConfiguration.hpp>
#include <SystemMngt/platforminterface/PlatformInterface.hpp>
#include <Repository/repositoryinterface/RepositoryInterface.hpp>

#include <iostream>

using namespace std;

// FT process related data.
static std::string m_application_name;
static std::string m_process_name;


Cdmw::clock::svcs::GroupsCreatorProcessBehaviour::GroupsCreatorProcessBehaviour(CORBA::ORB_ptr     orb,
                                                                                bool               point_to_point,
                                                                                const std::string& host1,
                                                                                const std::string& host2,
                                                                                const std::string& host3)
{
    m_orb              = CORBA::ORB::_duplicate(orb);
    m_point_to_point   = point_to_point;
    m_host1            = host1;
    m_host2            = host2; 
    m_host3            = host3;
}

Cdmw::clock::svcs::GroupsCreatorProcessBehaviour::~GroupsCreatorProcessBehaviour() throw ()
{
}

CORBA::ULong
Cdmw::clock::svcs::GroupsCreatorProcessBehaviour::nb_steps()
    throw (CORBA::SystemException)
{
    return 1;
}

void
Cdmw::clock::svcs::GroupsCreatorProcessBehaviour::on_initialise(const CdmwPlatformMngtBase::StartupKind&)
    throw (CORBA::SystemException)
{
    // Get application and process name
    m_application_name = Cdmw::PlatformMngt::PlatformInterface::Get_application_name();
    m_process_name     = Cdmw::PlatformMngt::PlatformInterface::Get_process_name();
    std::cout << "   -------- Server " << m_process_name << ": on_initialise [BEGIN] --------" << std::endl;

    // Get the Replication Manager
    std::cout << "       ---- Server " << m_process_name << ": get Replication Manager" << std::endl;
    CORBA::Object_var replication_mgr_obj = m_orb->resolve_initial_references("ReplicationManager");
    m_replication_mgr = CdmwFT::ReplicationManager::_narrow(replication_mgr_obj.in());

    // Get executor factory infos
    Cdmw::FT::FactoryInfos executor_factory_infos;
    GroupsCreatorProcessBehaviour::const_iterator executor_it_beg = executor_group_links.begin();
    GroupsCreatorProcessBehaviour::const_iterator executor_it_end = executor_group_links.end();
    GroupsCreatorProcessBehaviour::const_iterator executor_it_cur;
    for (executor_it_cur = executor_it_beg; executor_it_cur != executor_it_end ; ++executor_it_cur)
    {
        const std::string& process = executor_it_cur->first;
        const std::string& host    = executor_it_cur->second;
        if (process.empty() || host.empty())
        {
            std::cerr << "   -------- Server " << m_process_name << ": ERROR: process or host are empty" << std::endl;
            continue;
        }
        executor_factory_infos.add_ftLocation(host.c_str(), m_application_name.c_str(), process.c_str());
    }	

    // Build executor group properties
    ::FT::Properties executor_properties;
    executor_properties.length(4);
    executor_properties[0].nam.length(1);
    executor_properties[0].nam[0].id = "org.omg.ft.MinimumNumberReplicas";
    executor_properties[0].val     <<= (CORBA::UShort) 2;
    executor_properties[1].nam.length(1);
    executor_properties[1].nam[0].id = "org.omg.ft.ReplicationStyle";
    executor_properties[1].val     <<= (CORBA::UShort) 2;
    executor_properties[2].nam.length(1);
    executor_properties[2].nam[0].id = "org.omg.ft.MembershipStyle";
    executor_properties[2].val     <<= (CORBA::UShort) 0;
    executor_properties[3].nam.length(1);
    executor_properties[3].nam[0].id = "org.omg.ft.Factories";
    executor_properties[3].val     <<= executor_factory_infos.get_factoryInfos();

    // Build executor group criteria
    ::FT::Criteria executor_criteria;
    executor_criteria.length(2);
    executor_criteria[0].nam.length(1);
    executor_criteria[0].nam[0].id = "org.omg.ft.FTProperties";
    executor_criteria[0].val     <<= executor_properties;
    if (m_point_to_point)
    {
        // For point-to-point State Transfer, the identifiers of the used DataStores have to be added to the criteria
        std::cout << "       ---- Server " << m_process_name << ": add dsids into criteria" << std::endl;

        CdmwFT::StateTransfer::DataStoreIDs dsids;
        dsids.length(1);
        dsids[0] = CONTROLLER_DATA_IDENTIFIER;

        executor_criteria[1].nam.length(1);
        executor_criteria[1].nam[0].id = "com.thalesgroup.cdmw.ft.DataStoreIdentifiers";
        executor_criteria[1].val     <<= dsids;
    }

    // Create the executor object group
    ::FT::GenericFactory::FactoryCreationId_var factory_creation_id_for_executor_group;
    try
    {
        // Create the executor object group
        std::cout << "       ---- Server " << m_process_name << ": create the executor object group" << std::endl;
        m_executor_obj_group = m_replication_mgr->create_object(EXECUTOR_REPOSITORY_ID.c_str(), executor_criteria,
                                                                factory_creation_id_for_executor_group.out());
    }
    catch (CORBA::Exception& e)
    {
        std::cerr << __FILE__ << " " << __LINE__ << ": CORBA exception raised: " << e._name() << std::endl;
        throw;
    }


    if (!periodic_group_links.empty())
    {
        // Get periodic factory infos
        Cdmw::FT::FactoryInfos periodic_factory_infos;
        GroupsCreatorProcessBehaviour::const_iterator periodic_it_beg = periodic_group_links.begin();
        GroupsCreatorProcessBehaviour::const_iterator periodic_it_end = periodic_group_links.end();
        GroupsCreatorProcessBehaviour::const_iterator periodic_it_cur;
        for (periodic_it_cur = periodic_it_beg; periodic_it_cur != periodic_it_end ; ++periodic_it_cur)
        {
            const std::string& process = periodic_it_cur->first;
            const std::string& host    = periodic_it_cur->second;
            if (process.empty() || host.empty())
            {
                std::cerr << "   -------- Server " << m_process_name << ": ERROR: process or host are empty" << std::endl;
                continue;
            }
            periodic_factory_infos.add_ftLocation(host.c_str(), m_application_name.c_str(), process.c_str());
        }	

        // Build periodic group properties
        ::FT::Properties periodic_properties;
        periodic_properties.length(4);
        periodic_properties[0].nam.length(1);
        periodic_properties[0].nam[0].id = "org.omg.ft.MinimumNumberReplicas";
        periodic_properties[0].val     <<= (CORBA::UShort) 2;
        periodic_properties[1].nam.length(1);
        periodic_properties[1].nam[0].id = "org.omg.ft.ReplicationStyle";
        periodic_properties[1].val     <<= (CORBA::UShort) 2;
        periodic_properties[2].nam.length(1);
        periodic_properties[2].nam[0].id = "org.omg.ft.MembershipStyle";
        periodic_properties[2].val     <<= (CORBA::UShort) 0;
        periodic_properties[3].nam.length(1);
        periodic_properties[3].nam[0].id = "org.omg.ft.Factories";
        periodic_properties[3].val     <<= periodic_factory_infos.get_factoryInfos();

        // Build periodic group criteria
        ::FT::Criteria periodic_criteria;
        periodic_criteria.length(1);
        periodic_criteria[0].nam.length(1);
        periodic_criteria[0].nam[0].id = "org.omg.ft.FTProperties";
        periodic_criteria[0].val     <<= periodic_properties;

        // Create the periodic object group
        ::FT::GenericFactory::FactoryCreationId_var factory_creation_id_for_periodic_group;
        try
        {
            // Create the periodic object group
            std::cout << "       ---- Server " << m_process_name << ": create the periodic object group" << std::endl;
            m_periodic_obj_group = m_replication_mgr->create_object(PERIODIC_REPOSITORY_ID.c_str(), periodic_criteria, 
                                                                factory_creation_id_for_periodic_group.out());
        }
        catch (CORBA::Exception& e)
        {
            std::cerr << __FILE__ << " " << __LINE__ << ": CORBA exception raised: " << e._name() << std::endl;
            throw;
        }
    }

    // Get Repository
    std::cout << "       ---- Server " << m_process_name << ": get Repository" << std::endl;
    CdmwNamingAndRepository::Repository_var repository = Cdmw::NamingAndRepository::RepositoryInterface::get_repository();

    // Get NameDomain of object groups (for writing)
    m_obj_grps_domain = repository->resolve_name_domain(REPOSITORY_CLOCK_GROUP_NAME.c_str());

    // Register executor ObjectGroup into repository
    std::cout << "       ---- Server " << m_process_name << ": register executor object group into repository" << std::endl;
    CdmwNamingAndRepository::NameDomain::RegistrationId_var reg_id_for_executor_group = m_obj_grps_domain->new_name(EXECUTOR_GROUP_NAME.c_str());
    m_obj_grps_domain->register_object(reg_id_for_executor_group.in(), m_executor_obj_group.in());

    if (!periodic_group_links.empty())
    {
        // Register periodic ObjectGroup into repository
        std::cout << "       ---- Server " << m_process_name << ": register periodic object group into repository" << std::endl;
        CdmwNamingAndRepository::NameDomain::RegistrationId_var reg_id_for_periodic_group = m_obj_grps_domain->new_name(PERIODIC_GROUP_NAME.c_str());
        m_obj_grps_domain->register_object(reg_id_for_periodic_group.in(), m_periodic_obj_group.in());
    }

    std::cout << "   -------- Server " << m_process_name << ": on_initialise [ END ] --------" << std::endl;
}

void
Cdmw::clock::svcs::GroupsCreatorProcessBehaviour::on_run()
    throw (CdmwPlatformMngt::ProcessDelegate::NotReadyToRun,
           CORBA::SystemException)
{
    std::cout << "   -------- Server " << m_process_name << ": on_run [BEGIN] --------" << std::endl;
    std::cout << "   -------- Server " << m_process_name << ": on_run [ END ] --------" << std::endl;
}

void
Cdmw::clock::svcs::GroupsCreatorProcessBehaviour::on_stop()
    throw (CORBA::SystemException)
{
    std::cout << "   -------- Server " << m_process_name << ": on_stop [BEGIN] --------" << std::endl;
    m_orb->shutdown(true);
    std::cout << "   -------- Server " << m_process_name << ": on_stop [ END ] --------" << std::endl;
}

void
Cdmw::clock::svcs::GroupsCreatorProcessBehaviour::link_executor_process_and_host(const std::string& process, const std::string& host)
{
    executor_group_links.insert(make_pair(process, host));
}

void
Cdmw::clock::svcs::GroupsCreatorProcessBehaviour::link_periodic_process_and_host(const std::string& process, const std::string& host)
{
    periodic_group_links.insert(make_pair(process, host));
}
