/* ========================================================================== *
 * This file is part of CARDAMOM (R) which is jointly developed by THALES
 * and SELEX Sistemi Integrati.
 * Copyright (C) SELEX Sistemi Integrati 2004-2006. All rights reserved.
 *  
 * CARDAMOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * CARDAMOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public
 * License for more details.
 *  
 *  
 * You should have received a copy of the GNU Library General Public
 * License along with CARDAMOM; see the file COPYING. If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * ========================================================================= */

#include <Foundation/ossupport/OS.hpp>
#include <Foundation/orbsupport/OrbSupport.hpp>
#include <Foundation/orbsupport/StrategyList.hpp>
#include <FaultTolerance/ftinit/FTServiceInit.hpp>
#include <ConfAndPlug/cdmwinit/CdmwInterface.hpp>

#include "ftperiodic/FTPeriodicServiceProcessBehaviour.hpp"
#include "clockservice/StateTransferConfig.hpp"

static const int POA_THREAD_POOL_SIZE = 10;

int main(int argc, char* argv[])
{
    int ret_code = 0;
    CORBA::ORB_var orb;

    try
    {
        // Get host name
        std::string host_name = Cdmw::OsSupport::OS::get_hostname();

        // Check if primary option is specified
        std::string primary_option = Cdmw::OsSupport::OS::get_option_value(argc, argv, SET_PRIMARY_OPTION);

        bool set_as_primary=false;
        if (primary_option == "yes")
        {
            set_as_primary = true;
            std::cout << "       ---- " << host_name << " FT Periodic Service: this periodic will be primary" << std::endl;
        }

        // Initialise FT service (to be done before ORB_init)
        std::cout << "       ---- " << host_name << " FT Periodic Service: initialise FT service" << std::endl;
        Cdmw::CdmwInit::CDMW_preORBInit(argc, argv);

        // Set ORB strategies
        Cdmw::OrbSupport::StrategyList strategyList;
        strategyList.add_OrbThreaded();
        strategyList.add_PoaThreadPool(POA_THREAD_POOL_SIZE);

        // Initialise the ORB
        std::cout << "       ---- " << host_name << " FT Periodic Service: initialise the ORB" << std::endl;
        orb = Cdmw::OrbSupport::OrbSupport::ORB_init(argc, argv, strategyList);

        // Get the root POA
        std::cout << "       ---- " << host_name << " FT Periodic Service: get the root POA" << std::endl;
        CORBA::Object_var obj = orb->resolve_initial_references("RootPOA");
        PortableServer::POA_ptr rootPOA = PortableServer::POA::_narrow(obj.in());

        // Activate the POA manager
        std::cout << "       ---- " << host_name << " FT Periodic Service: activate the POA manager" << std::endl;
        PortableServer::POAManager_var poaMgr = rootPOA->the_POAManager();
        poaMgr->activate();

        // Creates the process behaviour
        std::cout << "       ---- " << host_name << " FT Periodic Service: create the process behaviour" << std::endl;
        typedef Cdmw::clock::svcs::FTPeriodicServiceProcessBehaviour PRCBehaviour;
        std::auto_ptr<PRCBehaviour> process_behaviour_auto_ptr(new PRCBehaviour(orb.in(), set_as_primary));

        // Initialize Cardamom
        std::cout << "       ---- " << host_name << " FT Periodic Service: initializing Cardamom" << std::endl;
        Cdmw::CdmwInit::CDMW_init(orb.in(), argc, argv, process_behaviour_auto_ptr.get());

        // Enter in the main loop
        std::cout << "       ---- " << host_name << " FT Periodic Service: listening for CORBA requests" << std::endl;
        orb->run();
    }
    catch (const CORBA::Exception &e)
    {
        std::cerr << __FILE__ << " " << __LINE__ << ": CORBA::Exception exception raised: " << e._name() << std::endl;
        ret_code = EXIT_FAILURE;
    }

    if (!CORBA::is_nil(orb.in()))
    {
        try
        {
            Cdmw::CdmwInit::CDMW_cleanup(orb.in());
            Cdmw::OrbSupport::OrbSupport::ORB_cleanup(orb.in());
            orb->destroy();
        }
        catch (const CORBA::Exception& e)
        {
            std::cerr << __FILE__ << " " << __LINE__ << ": CORBA::Exception exception raised: " << e._name() << std::endl;
            ret_code = EXIT_FAILURE;
        }
    }

    // Exit the main program
    return ret_code;
}
