/* =========================================================================== *
 * This file is part of CARDAMOM (R) which is jointly developed by THALES
 * and SELEX Sistemi Integrati.
 * Copyright (C) SELEX Sistemi Integrati 2004-2006. All rights reserved.
 *  
 * CARDAMOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * CARDAMOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public
 * License for more details.
 *  
 *  
 * You should have received a copy of the GNU Library General Public
 * License along with CARDAMOM; see the file COPYING. If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * =========================================================================== */

// Cdmw Files
#include <Foundation/orbsupport/CORBA.hpp>
#include <Foundation/ossupport/OS.hpp>
#include <SystemMngt/platforminterface/PlatformInterface.hpp>
#include <Foundation/commonsvcs/naming/NamingUtil.hpp>
#include <Repository/repositoryinterface/RepositoryInterface.hpp>
#include <FaultTolerance/idllib/CdmwFTReplicationManager.stub.hpp>
#include <FaultTolerance/ftreplicationmanager/CdmwFTObjectGroupData.stub.hpp>

// Demo Files
#include "ftperiodic/FTPeriodicServiceProcessBehaviour.hpp"
#include "clockservice/StateTransferConfig.hpp"

/*
#include "ace/Thread.h"
#include "ace/Thread_Manager.h"
#include "ace/TP_Reactor.h"
#include "ace/Select_Reactor.h"
#include "ace/Get_Opt.h"
#include "ace/Task.h"
#include "ace/Timer_Queue.h"
#include "ace/Timer_Heap.h"
#include "ace/Timer_Wheel.h"
#include "ace/Timer_Hash.h"
#include "ace/Timer_List.h"
*/

// FT process related data.
static std::string m_application_name;
static std::string m_process_name;
static std::string m_location;


Cdmw::clock::svcs::FTPeriodicServiceProcessBehaviour::
FTPeriodicServiceProcessBehaviour(CORBA::ORB_ptr orb, bool primary)
    throw ()
{
    m_orb = CORBA::ORB::_duplicate(orb);
    m_set_has_primary = primary;
}

Cdmw::clock::svcs::FTPeriodicServiceProcessBehaviour::~FTPeriodicServiceProcessBehaviour()
    throw ()
{
}

CORBA::ULong
Cdmw::clock::svcs::FTPeriodicServiceProcessBehaviour::nb_steps()
    throw (CORBA::SystemException)
{
    return 1;
}

void
Cdmw::clock::svcs::FTPeriodicServiceProcessBehaviour::on_initialise(const CdmwPlatformMngtBase::StartupKind&)
    throw (CORBA::SystemException)
{
    // Get application and process name
    m_application_name = Cdmw::PlatformMngt::PlatformInterface::Get_application_name();
    m_process_name     = Cdmw::PlatformMngt::PlatformInterface::Get_process_name();
    std::cout << "   -------- Server " << m_process_name << ": on_initialise [BEGIN] --------" << std::endl;

    // Get root POA
    std::cout << "   -------- Server " << m_process_name << ": get the root POA" << std::endl;
    CORBA::Object_var obj = m_orb->resolve_initial_references("RootPOA");
    PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(obj.in());
    
    // Get the POA manager
    std::cout << "   -------- Server " << m_process_name << ": get the POA manager" << std::endl;
    PortableServer::POAManager_var poaMgr = rootPOA->the_POAManager();

    // Create a POA with USE_SERVANT_MANAGER & NON_RETAIN policy (to use ServantLocator)
    CORBA::PolicyList policies;
    policies.length(3);
    policies[0] = rootPOA->create_request_processing_policy(PortableServer::USE_SERVANT_MANAGER);
    policies[1] = rootPOA->create_servant_retention_policy(PortableServer::NON_RETAIN);
    policies[2] = rootPOA->create_id_assignment_policy(PortableServer::USER_ID);
    Cdmw::OrbSupport::StrategyList strategyList;
    std::cout << "   -------- Server " << m_process_name << ": create dedicated POA" << std::endl;
    PortableServer::POA_var myPOA = Cdmw::OrbSupport::OrbSupport::create_POA(rootPOA.in(),
                                                                             "myPOA",
                                                                             poaMgr.in(),
                                                                             policies,
                                                                             strategyList);

    // Get executor and periodic group references from Cardamom Repository
    CosClockService::PeriodicExecution::Executor_var executor_group;
    CosClockService::PeriodicExecution::Periodic_var periodic_group;
    get_executor_and_periodic_groups(executor_group, periodic_group);

    // Create periodic servant
    std::cout << "       ---- Server " << m_process_name << ": create periodic servant" << std::endl;
    m_periodic_impl = new Cdmw::clock::FTPeriodic_impl(executor_group.in());

    // Create periodic object under myPOA
    std::cout << "       ---- Server " << m_process_name << ": create periodic reference " << std::endl;
    std::string periodic_oid = m_process_name + std::string("_periodic_oid");
    PortableServer::ObjectId_var periodic_obj_id = PortableServer::string_to_ObjectId(periodic_oid.c_str());
    CORBA::Object_var periodic_obj = myPOA->create_reference_with_id(periodic_obj_id, "IDL:omg.org/CosClockService/PeriodicExecution/Periodic:1.0");
    m_periodic_obj = CosClockService::PeriodicExecution::Periodic::_narrow(periodic_obj.in());

    // Create periodic activation handler
    std::cout << "       ---- Server " << m_process_name << ": create periodic activation handler" << std::endl;
    std::string periodic_activation_handler_name = m_process_name + std::string("_periodic_activation_handler_name");
    m_periodic_activation_handler = new Cdmw::clock::FTPeriodicActivationHandler(periodic_activation_handler_name, m_periodic_impl);

    // Create periodic servant locator
    std::cout << "       ---- Server " << m_process_name << ": create servant locator" << std::endl;
    m_periodic_servant_locator = new Cdmw::clock::FTPeriodicServantLocator(periodic_oid, m_periodic_activation_handler);

    // Set periodic servant locator to myPOA
    try
    {
        myPOA->set_servant_manager(m_periodic_servant_locator);
    }
    catch (const PortableServer::POA::WrongPolicy& e)
    {
        // Shouldn't happen!!
        CDMW_ASSERT(false);
    }

    // Add periodic object as member of periodic group.
    this->add_periodic_to_periodic_group(periodic_group);

    // Activate POAManager
    std::cout << "       ---- Server " << m_process_name << ": activate POAManager " << std::endl;
    poaMgr->activate();

    std::cout << "   -------- Server " << m_process_name << ": on_initialise [ END ] --------" << std::endl;
}

void Cdmw::clock::svcs::FTPeriodicServiceProcessBehaviour::on_next_step()
    throw( CdmwPlatformMngt::ProcessDelegate::InvalidStep,
           CORBA::SystemException )
{
}

void Cdmw::clock::svcs::FTPeriodicServiceProcessBehaviour::on_run()
    throw( CdmwPlatformMngt::ProcessDelegate::NotReadyToRun,
           CORBA::SystemException )
{
    std::cout << "   -------- Server " << m_process_name << ": on_run [BEGIN] --------" << std::endl;
    std::cout << "   -------- Server " << m_process_name << ": on_run [ END ] --------" << std::endl;
}

void Cdmw::clock::svcs::FTPeriodicServiceProcessBehaviour::on_stop()
    throw( CORBA::SystemException )
{
    std::cout << "   -------- Server " << m_process_name << ": on_stop [BEGIN] --------" << std::endl;
    m_orb->shutdown(true);
    std::cout << "   -------- Server " << m_process_name << ": on_stop [ END ] --------" << std::endl;
}

/**
 * Add Periodic object as a new member of the periodic group.
 */
void
Cdmw::clock::svcs::FTPeriodicServiceProcessBehaviour::
add_periodic_to_periodic_group(const CosClockService::PeriodicExecution::Periodic_var& periodic_group)
{
    // Get the group repository
    try
    {
        std::cout << "   -------- Server " << m_process_name << ": get the FT group repository" << std::endl;
        CORBA::Object_var obj = m_orb->resolve_initial_references("FTGroupRepository");
        m_group_repository = CdmwFT::Location::PrimaryBackupGroupRepository::_narrow(obj.in());
    }
    catch (const CORBA::ORB::InvalidName&)
    {
        std::cerr << __FILE__ << " " << __LINE__ << ": CORBA::ORB::InvalidName exception raised" << std::endl;
    }

    // Get this process' location
    std::auto_ptr< ::FT::Location > the_location(m_group_repository->the_location());
    m_location = Cdmw::CommonSvcs::Naming::NamingInterface::to_string(*(the_location.get()));
    std::cout << "       ---- Server " << m_process_name << ": location is " << m_location << std::endl;

    // Get the replication manager
    std::cout << "       ---- Server " << m_process_name << ": get the Replication Manager" << std::endl;
    CORBA::Object_var replication_mgr_obj = m_orb->resolve_initial_references("ReplicationManager");
    m_replication_mgr = CdmwFT::ReplicationManager::_narrow(replication_mgr_obj.in());

    // Create location for this member
    std::cout << "       ---- Server " << m_process_name << ": create Location for this member" << std::endl;
    ::FT::Location loc;
    loc.length(3);
    loc[0].id   = CORBA::string_dup(Cdmw::OsSupport::OS::get_hostname().c_str());
    loc[0].kind = "hostname";
    loc[1].id   = m_application_name.c_str();
    loc[1].kind = "applicationname";
    loc[2].id   = m_process_name.c_str();
    loc[2].kind = "processname";

    // Add periodic object as a new member of the periodic object group
    ::FT::ObjectGroup_var periodic_obj_group;
    try
    {
        // Add periodic object as a new member of the periodic object group
        std::cout << "       ---- Server " << m_process_name << ": add " << m_process_name << " to periodic object group" << std::endl;
        periodic_obj_group = m_replication_mgr->add_member(periodic_group.in(), loc, m_periodic_obj.in());
    }
    catch (const ::FT::ObjectGroupNotFound&)
    {
        std::cerr << __FILE__ << "  " << __LINE__ << ": FT::ObjectGroupNotFound exception raised" << std::endl;
        return;
    }
    catch (const ::FT::MemberAlreadyPresent&)
    {
        std::cerr << __FILE__ << "  " << __LINE__ << ": FT::MemberAlreadyPresent exception raised" << std::endl;
        return;
    }
    catch (const ::FT::ObjectNotAdded&)
    {
        std::cerr << __FILE__ << "  " << __LINE__ << ": FT::ObjectNotAdded exception raised" << std::endl;
        return;
    }

    // Register periodic activation handler
    std::cout << "       ---- Server " << m_process_name << ": register periodic activation handler" << std::endl;
    try
    {
        // Get ActivationManager
        std::cout << "       ---- Server " << m_process_name << ": get the FT Activation Manager" << std::endl;
        CORBA::Object_var obj = m_orb->resolve_initial_references("FTActivationManager");
        CdmwFT::Location::ActivationManager_var activation_manager = CdmwFT::Location::ActivationManager::_narrow(obj.in());

        // Register activation handler for periodic group into ActivationManager
        std::cout << "       ---- Server " << m_process_name << ": register activation handler for periodic group" << std::endl;
        activation_manager->register_activation_handler(periodic_group.in(), m_periodic_activation_handler);
    }
    catch (const ::FT::ObjectGroupNotFound&)
    {
        std::cerr << __FILE__ << "  " << __LINE__ << ": FT::ObjectGroupNotFound exception raised" << std::endl;
    }
    catch (const ::CORBA::ORB::InvalidName&)
    {
        std::cerr << __FILE__ << " " << __LINE__ << ": CORBA::ORB::InvalidName exception raised" << std::endl;
    }
    catch (const ::CORBA::SystemException& e)
    {
        std::cerr << __FILE__ << " " << __LINE__ << ": CORBA::SystemException exception raised: " << e._name() << std::endl;
    }

    // Set periodic as primary if required
    if (m_set_has_primary)
    {
        try
        {
            // Set periodic as primary
            std::cout << "       ---- Server " << m_process_name << ": set " << m_process_name << " as periodic primary" << std::endl;
            periodic_obj_group = m_replication_mgr->set_primary_member(periodic_obj_group.in(), loc);
        }
        catch (const ::FT::ObjectGroupNotFound& e)
        {
            std::cerr << __FILE__ << "  " << __LINE__ << ": FT::ObjectGroupNotFound exception raised" << std::endl;
        }
        catch (const ::FT::MemberNotFound& e)
        {
            std::cerr << __FILE__ << "  " << __LINE__ << ": FT::MemberNotFound exception raised" << std::endl;
        }
        catch (const ::FT::PrimaryNotSet& e)
        {
            std::cerr << __FILE__ << "  " << __LINE__ << ": FT::PrimaryNotSet exception raised" << std::endl;
        }
        catch (const ::FT::BadReplicationStyle& e)
        {
            std::cerr << __FILE__ << "  " << __LINE__ << ": FT::BadReplicationStyle exception raised" << std::endl;
        }
    }
}

/**
 * Get Executor group reference from Cardamom Repository.
 * Get Periodic group reference from Cardamom Repository.
 * These references were created by the groups_creator process.
 */
void
Cdmw::clock::svcs::FTPeriodicServiceProcessBehaviour::
get_executor_and_periodic_groups(CosClockService::PeriodicExecution::Executor_var& executor_group,
                                 CosClockService::PeriodicExecution::Periodic_var& periodic_group)
{
    // Get the Cardamom Repository
    CdmwNamingAndRepository::Repository_var repository =
        Cdmw::NamingAndRepository::RepositoryInterface::get_repository();

    // Get NamingInterface to executor and periodic object groups (for reading)
    Cdmw::CommonSvcs::Naming::NamingInterface obj_grps_interface =
        Cdmw::NamingAndRepository::RepositoryInterface::get_domain_naming_interface(REPOSITORY_CLOCK_GROUP_NAME.c_str());

    // Get executor and periodic Object Group
    CosClockService::PeriodicExecution::Executor_var executor_grp = CosClockService::PeriodicExecution::Executor::_nil();
    CosClockService::PeriodicExecution::Periodic_var periodic_grp = CosClockService::PeriodicExecution::Periodic::_nil();
    try
    {
        // Get executor object group
        std::cout << "       ---- Server " << m_process_name << ": get Executor object group " << std::endl;
        CORBA::Object_var executor_obj = obj_grps_interface.resolve(EXECUTOR_GROUP_NAME);
        executor_grp = CosClockService::PeriodicExecution::Executor::_narrow(executor_obj.in());

        // Get periodic object group
        std::cout << "       ---- Server " << m_process_name << ": get Periodic object group " << std::endl;
        CORBA::Object_var periodic_obj = obj_grps_interface.resolve(PERIODIC_GROUP_NAME);
        periodic_grp = CosClockService::PeriodicExecution::Periodic::_narrow(periodic_obj.in());

        // Assign the computed groups to the output variables
        executor_group = executor_grp;
        periodic_group = periodic_grp;
    }
    catch (const CosNaming::NamingContext::NotFound& e)
    {
            std::cerr << __FILE__ << "  " << __LINE__ << ": CosNaming::NamingContext::NotFound exception raised" << std::endl;
    }
    catch (const CosNaming::NamingContext::CannotProceed& e)
    {
            std::cerr << __FILE__ << "  " << __LINE__ << ": CosNaming::NamingContext::CannotProceed exception raised" << std::endl;
    }
    catch (const CosNaming::NamingContext::InvalidName& e)
    {
            std::cerr << __FILE__ << "  " << __LINE__ << ": CosNaming::NamingContext::InvalidName exception raised" << std::endl;
    }
    catch (const CORBA::SystemException& e)
    {
            std::cerr << __FILE__ << "  " << __LINE__ << ": CORBA::SystemException exception raised: " << e._name() << std::endl;
    }
}
