/* ===================================================================== */
/*
 * This file is part of CARDAMOM (R) which is jointly developed by THALES
 * and SELEX-SI. It is derivative work based on PERCO Copyright (C) THALES 
 * 2000-2003. All rights reserved.
 * Copyright (C) THALES 2004-2005. All rights reserved.
 * Copyright (C) SELEX-SI and THALES 2006. All rights reserved.
 * 
 * CARDAMOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * CARDAMOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with CARDAMOM; see the file COPYING. If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* ===================================================================== */


#include <iostream>
#include <sstream>

#include "tracecollector/LocalCollectorMngr.hpp"
#include "tracecollector/TraceCircularBuffer.hpp"

#include "TraceAndPerf/idllib/CdmwTraceTraceProducer.stub.hpp"
#include "TraceAndPerf/tracelibrary/FlushAreaMngr.hpp"

namespace Cdmw
{

namespace Trace
{


/**
 * Constructor.
 *
 */
LocalCollectorMngr::LocalCollectorMngr(
        CORBA::ORB_ptr orb,
        PortableServer::POA_ptr trace_POA,
        LocalTraceCollector_impl::CollectorDataList& global_collectors,
        CollectorInit::CollectorOptions& options,
        OsSupport::OS::TimeMode time_mode,
        CdmwTrace::Collector_ptr miop_collectors)
    throw(OutOfMemoryException,
          BadParameterException,
          InternalErrorException)
        : CollectorMngr(trace_POA, 
                        options.component_name, 
                        options.domain_name, 
                        options.level_value,
                        options.trace_format, 
                        options.queue_strategy, 
                        options.queue_size,
                        options.queue_str_size, 
                        options.trace_file_name, 
                        options.trace_file_bak_num,
                        options.trace_file_num, 
                        options.trace_msg_num, 
                        time_mode)
{
    // We must create the Servant to handle registration request (ref count is incremented)
    LocalTraceCollector_impl* p_local_collector_servant =
                             new LocalTraceCollector_impl (this, m_filter_mngr);

    // create an object var to take pointer ownership
    // (ref count will be decremented when var will be destroyed at the method end)
    PortableServer::ServantBase_var servant_var = p_local_collector_servant;

    // create an object id
    PortableServer::ObjectId_var oid =
                  PortableServer::string_to_ObjectId("TraceCollector");
    try
    {
        // activate servant on POA (ref count is incremented)
        trace_POA->activate_object_with_id(oid.in(), p_local_collector_servant);
        CORBA::Object_var object = trace_POA->id_to_reference(oid.in());
        m_trace_collector = CdmwTrace::Collector::_narrow(object.in());
    }
    catch (const PortableServer::POA::WrongPolicy &)
    {
        CDMW_THROW2(Cdmw::BadParameterException, "Wrong POA Policy for activate_object", "NULL");
    }
    catch (const PortableServer::POA::ObjectAlreadyActive & )
    {
        CDMW_THROW2(BadParameterException, "Trace Collector Object already active", "NULL"); 
    } 
    catch (const PortableServer::POA::ServantAlreadyActive & )
    {
        CDMW_THROW2(BadParameterException, "Trace Collector Servant already active", "NULL");  
    }
    catch (const CORBA::Exception& ex)
    {
        std::cerr << "Trace Collector servant activation error" << "\n";
        std::cerr << "CORBA Exception : \n" << ex << std::endl;
        CDMW_ASSERT(false);
    }
        
    if ( CORBA::is_nil(miop_collectors) )
    {
      // FlushAreaMngr needs POA with System ID
      CORBA::Object_var obj = orb->resolve_initial_references("RootPOA");
      PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(obj.in());

      CdmwTrace::TraceProducer_var dummyProducer =
          FlushAreaMngr::init (rootPOA.in(),
                               global_collectors,
                               "CollectorApplication",
                               "LocalCollector",
                               2000, // 2 seconds
                               Cdmw::Trace::FlushAreaMngr::DEFAULT_AREA_NB,
                       // Size of flush area to take into account serialization
                               options.size_flush_area,
                       // number of messages for each area
                               options.area_nb_messages,
                       // threshold for repetitive messages
                               Cdmw::Trace::FlushAreaMngr::DEFAULT_MSG_THRESHOLD,
                               orb
          );

      // activate the flushing
      FlushAreaMngr& flushAreaMngr=FlushAreaMngr::instance();
      flushAreaMngr.activate_flushing(); 

      // register the global collectors to resend messages
      p_local_collector_servant->set_global_collectors(global_collectors, orb);
    }
    else
    {
      p_local_collector_servant->set_miop_collectors(miop_collectors);
    }
}




/**
 * Initialize the Trace collector.
 */
CdmwTrace::Collector_ptr
LocalCollectorMngr::Init(
        CORBA::ORB_ptr orb,
        PortableServer::POA_ptr trace_POA,
        LocalTraceCollector_impl::CollectorDataList& global_collectors,
        CollectorInit::CollectorOptions& options,
        OsSupport::OS::TimeMode time_mode,
        CdmwTrace::Collector_ptr miop_collectors)
    throw(OutOfMemoryException,
          BadParameterException,
          InternalErrorException)
{
    // The user try to initialise twice the collector
    CDMW_ASSERT (M_init_done == false);

    try 
    {
        M_singleton = new  LocalCollectorMngr (orb,
                                               trace_POA,
                                               global_collectors,
                                               options,
                                               time_mode,
                                               miop_collectors);
    }
    catch (const std::bad_alloc&) 
    {
        CDMW_THROW (OutOfMemoryException);
    }

    M_init_done = true;

    // We return the reference on the object managing communication with
    //  collector, it will be registered in the repository by the caller
    return M_singleton->get_trace_collector();
}



//
// Dtor
// 
LocalCollectorMngr::~LocalCollectorMngr()
    throw()
{
}


// 
// write message in the circular buffer to be flushed on disk
//
bool LocalCollectorMngr::write_message (const CdmwTrace::MessagesHeader& messages_hdr,
                                        const CdmwTrace::FormattedMessage& fmt_message)
     throw (OutOfMemoryException)
{
    bool ret = true;

    CDMW_ASSERT (m_collector_flusher.get() != NULL);
    CDMW_ASSERT (m_circular_buffer != NULL);

    // write message in circular buffer
    // collector flusher is signaled to flush message on disk
    TraceCircularBuffer::WriteStatus write_status;
    ret = m_circular_buffer->writeMessage (messages_hdr, fmt_message, write_status);


    if  (write_status == TraceCircularBuffer::STS_DROP_LATEST)
    {
        m_collector_flusher->print_error_message(
                        "Queue is full : latest trace message has been dropped");
    }
    else if  (write_status == TraceCircularBuffer::STS_REMOVE_OLDEST)
    {
        m_collector_flusher->print_error_message(
                        "Queue is full : oldest trace message has been removed");
    }
      
    return ret;
}



}; // End namespace Trace
}; // End namespace Cdmw

