/* ===================================================================== */
/*
 * This file is part of CARDAMOM (R) which is jointly developed by THALES
 * and SELEX-SI. It is derivative work based on PERCO Copyright (C) THALES 
 * 2000-2003. All rights reserved.
 * Copyright (C) THALES 2004-2005. All rights reserved.
 * Copyright (C) SELEX-SI and THALES 2006. All rights reserved.
 * 
 * CARDAMOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * CARDAMOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with CARDAMOM; see the file COPYING. If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* ===================================================================== */

#include "Foundation/ossupport/OS.hpp"
#include <Foundation/orbsupport/ExceptionMinorCodes.hpp>

#include "TraceAndPerf/tracelibrary/FlushAreaMngr.hpp"
#include "tracecollector/LocalCollectorMngr.hpp"
#include "tracecollector/LocalTraceCollector_impl.hpp"


namespace Cdmw
{
namespace Trace
{


/**
 * Constructor.
 */
LocalTraceCollector_impl::LocalTraceCollector_impl (CollectorMngr* p_collector_mngr,
                                                    Logging::FilterMngr* p_filter_mngr)
    throw()
        : TraceCollector_impl (p_collector_mngr,
                               p_filter_mngr)
{
  m_miop_collectors = CdmwTrace::Collector::_nil();
  m_codec = 0;
}


/**
 * Destructor.
 */
LocalTraceCollector_impl::~LocalTraceCollector_impl()
    throw()
{
  if (m_codec != 0) delete m_codec;
}


/**
 * Write messages.
 */
void
LocalTraceCollector_impl::put_formatted_messages(
        const CdmwTrace::MessagesHeader& the_messages_header,
        const CdmwTrace::FormattedMessageSeq& the_messages)
    throw(CORBA::SystemException)
{
    try 
    {
        // write the message locally.
        for (size_t i_msg = 0; i_msg < the_messages.length(); ++i_msg) 
        {
            m_collector_mngr->write_message (the_messages_header,
                                             the_messages[i_msg]);
        }

        // resend the messages to the global collectors.
        
        // filter the messages to send to the global trace collector
        CdmwTrace::FormattedMessageSeq_var messages_to_transfer =
                   new CdmwTrace::FormattedMessageSeq();
                   
        unsigned int inx_transf_message = 0;
        
        // for each message to transfer
        for (size_t i_msg = 0; i_msg < the_messages.length(); ++i_msg) 
        {
            // write out in the sequence the message only 
            // if it passes at least one filter
            
            const CdmwTrace::FormattedMessage& fmt_message = the_messages[i_msg];
            std::string domain = fmt_message.header.the_level.the_domain.in();

            std::string component_name =
                 fmt_message.header.the_level.the_component_name.in();
                 
            if (CollectorMngr::Is_to_be_traced(
                                       component_name, domain,
                                       fmt_message.header.the_level.the_value))
            {
                inx_transf_message++;
                messages_to_transfer->length(inx_transf_message);
                messages_to_transfer[inx_transf_message-1] = fmt_message;
            }
        }
        
        // transfer messages if any
        if (messages_to_transfer->length() > 0)
        {
            if ( CORBA::is_nil(m_miop_collectors.in()) )
                //if ( m_miop_collectors.in() == CdmwTrace::Collector::_nil() )
            {
                // constructs a message header
                Logging::LogHeader header;
                header.m_fileLine = __LINE__;
                header.m_fileName = __FILE__;
                header.m_userLevel = Cdmw::Logging::INFO;
                header.m_userDomain = "LOCALCOLLECTOR";
                header.m_componentName = "";
                header.m_tid = OsSupport::Thread::self();
                header.m_timestamping = OsSupport::OS::get_time();

                CdmwTrace::MessagesWithHeader_var collectorMsg =
                  new CdmwTrace::MessagesWithHeader();
                collectorMsg->header = the_messages_header;
                collectorMsg->message_sequence = messages_to_transfer;
                // encode data
                CORBA::OctetSeq_var encodedData = m_codec->encode(collectorMsg.in());

                // add serialised message in the buffer
                FlushAreaMngr& flushAreaMngr = FlushAreaMngr::instance();
                if (! flushAreaMngr.add_serialized_message (header, encodedData.in()) )
                {
                    std::cerr <<
                      "Warning: message was not added in LocalTraceCollector buffer"
                      << std::endl;
                }

                /* For debug
                std::cout << "Local Trace collector, pid is:" << FlushAreaMngr::instance().getProcessId() << std::endl;
		std::cout << "LocalCollector put_formatted_messages, size:" << encodedData->length() << " seq length was:" << messages_to_transfer->length() << std::endl;
		*/

           }
            else
            {

	        //std::cout << "Send MIOP put_formatted_messages, seq length is:" << messages_to_transfer->length() << std::endl;
	        try {
		  // It is not mandatory to use bufferization for miop
		  // as it is "acceptable" to loose messages in miop
		  // advantage is to not transfer this part and parameters in tracelibrary
                  m_miop_collectors.in()->put_formatted_messages(
                                            the_messages_header,
                                            messages_to_transfer.in());
		}
                catch (const CORBA::Exception& ex)
                {
		  std::cerr << "Exception raised in Local Collector put_formated_message " << Cdmw::OrbSupport::OrbSupport::exception_to_string(ex) << std::endl;
                  throw;
                }

            } 
        }
    }
    catch(const OutOfMemoryException&) 
    {
        throw CORBA::NO_RESOURCES(OrbSupport::NO_RESOURCES,
                                  CORBA::COMPLETED_NO);
    }
}



/**
 * Set the global trace collectors.
 */
void
LocalTraceCollector_impl::set_global_collectors (
    CollectorDataList& collector_list,
    CORBA::ORB_ptr orb
    )
{
    m_global_collectors = collector_list;
    m_codec = new Cdmw::OrbSupport::Codec<CdmwTrace::MessagesWithHeader> (orb);
}

/**
 * Set the global trace collectors miop reference
 */
void
LocalTraceCollector_impl::set_miop_collectors (CdmwTrace::Collector_ptr miop_collectors)
{
    m_miop_collectors = CdmwTrace::Collector::_duplicate(miop_collectors);
}


} // namespace Trace
} // namespace Cdmw

