/* ===================================================================== */
/*
 * This file is part of CARDAMOM (R) which is jointly developed by THALES
 * and SELEX-SI. It is derivative work based on PERCO Copyright (C) THALES 
 * 2000-2003. All rights reserved.
 * Copyright (C) THALES 2004-2005. All rights reserved.
 * Copyright (C) SELEX-SI and THALES 2006. All rights reserved.
 * 
 * CARDAMOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * CARDAMOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with CARDAMOM; see the file COPYING. If not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* ===================================================================== */

#include <clocale>
#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>

#include <Foundation/common/Assert.hpp>
#include <Foundation/common/Exception.hpp>
#include <Foundation/common/Locations.hpp>
#include <Foundation/common/Options.hpp>
#include <Foundation/common/System.hpp>
#include <Foundation/commonsvcs/naming/NamingInterface.hpp>
#include <Foundation/commonsvcs/naming/NamingUtil.hpp>
#include <Foundation/ossupport/OS.hpp>
#include <Foundation/orbsupport/CORBA.hpp>
#include <Foundation/orbsupport/ExceptionMinorCodes.hpp>
#include <Foundation/orbsupport/OrbSupport.hpp>
#include <Foundation/orbsupport/StrategyList.hpp>

#include <SystemMngt/platforminterface/PlatformInterface.hpp>
#include <Repository/repositoryinterface/RepositoryInterface.hpp>

#include "TraceAndPerf/idllib/CdmwTraceTraceProducer.stub.hpp"

#include "tracecollector/CollectorInit.hpp"
#include "tracecollector/LocalCollectorMngr.hpp"


namespace
{
    using std::string;
    
    // default port number used by the local collector
    const std::string DFLT_SERVICE_PORT = "51234";
    const std::string DFLT_PROCESS_PORT = "51235";
    
    const string GLOBAL_COLLECTOR_INI_MNEMONAME = "global_collector_ini_";
    
    // ORB reference
    CORBA::ORB_var orb = CORBA::ORB::_nil();
    
}; // anonymous namespace

using namespace Cdmw;
using namespace Cdmw::OsSupport;
using namespace Cdmw::OrbSupport;
using namespace Cdmw::Trace;
using namespace Cdmw::PlatformMngt;


/**
 * Local Collector initialization:
 *
 * - create the POA for the collector servant under RootPOA.
 * - create the CollectorMngr object:
 *       + the collector servant is created and activated on its POA.
 *       + the flushing data background thread is started.
 * - export the collector servant to the IOR file.
 * - create ProcessAgentImpl object: the process agent servant is created
 *   and activated on the RootPOA.
 * - activate the POA Manager.
 * - retrieve the Naming & Repository service.
 * - register the collector servant in Naming & Repository.
 * - register the process agent servant as URL under corbaloc.
 */
int
init_local_collector (CORBA::ORB_ptr orb, 
                      PortableServer::POA_ptr root_POA,
                      CollectorInit::CollectorOptions& options,
                      CdmwNamingAndRepository::Repository_ptr p_repository_ref,
                      const std::string &trace_producer_context,
                      int argc, char* argv[])
{
    int status = FAILURE;

    using Cdmw::Common::Locations;

    // Collector servant reference
    CdmwTrace::Collector_var collector_ref = CdmwTrace::Collector::_nil();

    try 
    {
        // look for the global collectors (using naming interface).
        // the global collectors are registered in the following name domain:
        // CDMW/SERVICES/TRACE/COLLECTORS/
        std::string collector_domain_str(Locations::CDMW_SERVICES_NAME_DOMAIN);
        collector_domain_str += "/";
        collector_domain_str += Locations::TRACE_COLLECTORS_NAME_DOMAIN;

        LocalTraceCollector_impl::CollectorDataList global_collectors;
        
      // if we use MIOP no need to have each collector reference
      if ( options.miop_corbaloc == "" )
      {
        try 
        {
            // check if the global collector domain name is a registered name domain.
            CdmwNamingAndRepository::NameDomain_var collector_domain =
                p_repository_ref->resolve_name_domain(collector_domain_str.c_str());

            // get the default root context.
            CosNaming::NamingContext_var nc_root_context =
                p_repository_ref->resolve_root_context(
                                     CdmwNamingAndRepository::DEFAULT_ROOT_CONTEXT);

            using namespace Cdmw::CommonSvcs::Naming;
            
            // naming interface of default root context
            NamingInterface ni_root(nc_root_context.in());
            
            CosNaming::NamingContext_var nc_collector_dom =
                                     CosNaming::NamingContext::_nil();

            // get the global collector domain name.
            typedef NamingUtil<CosNaming::NamingContext> Util;
            nc_collector_dom = Util::resolve_name(ni_root,
                                                  collector_domain_str);

            CosNaming::BindingIterator_var binding_itr;
            CosNaming::BindingList_var binding_list;

            // naming interface of global collector domain
            NamingInterface ni_collector(nc_collector_dom.in());

            // get all the global collectors registered in this domain.
            ni_collector.list(NamingInterface::LIST_CHUNK_SIZE,
                              binding_list,
                              binding_itr);

            bool hasMoreBinding = false;
            int collector_index = 0;

            do 
            {
                for (size_t i_binding = 0;
                     i_binding < binding_list->length();
                     ++i_binding)
                {
                    CosNaming::Binding binding = binding_list[i_binding];

                    // get the collector name from the binding.
                    std::string glob_collector_name =
                        ni_collector.to_string(binding.binding_name);

                    // get the collector reference from its name
                    CORBA::Object_var obj = ni_collector.resolve(glob_collector_name);
                    CdmwTrace::Collector_var glob_collector =
                                             CdmwTrace::Collector::_narrow(obj.in());

                    if (!CORBA::is_nil(glob_collector.in())) 
                    {
                        //std::cout << "found " << glob_collector_name <<
                        //  " in repository" << std::endl;

                        // create a mnemonic name for the collector.
                        std::ostringstream mnemo_name_str;
                        mnemo_name_str << GLOBAL_COLLECTOR_INI_MNEMONAME
                                       << collector_index;
                        collector_index++;

                        // create a CollectorData structure.
                        CdmwTrace::TraceProducer::CollectorData glob_collector_data;
                        glob_collector_data.the_collectorMnemoName =
                            CORBA::string_dup(mnemo_name_str.str().c_str());
                        glob_collector_data.the_collectorObjName =
                            CORBA::string_dup(glob_collector_name.c_str());
                        glob_collector_data.the_collector = glob_collector;

                        // append it to the global collector list.
                        global_collectors.insert(global_collectors.end(),
                                                 glob_collector_data);
                    }
                }

                // should we continue?
                if (!CORBA::is_nil(binding_itr.in())) 
                {
                    hasMoreBinding =
                        binding_itr->next_n (NamingInterface::LIST_CHUNK_SIZE,
                                             binding_list);
                }
            } while (hasMoreBinding);
        }
        catch (const CosNaming::NamingContext::NotFound&) {
            // should not happen!
            std::cerr << "Trace Collector Name Domain not found \n"
                      << collector_domain_str.c_str()
                      << std::endl;
        }
        catch (const CdmwNamingAndRepository::Repository::NoRootContext&) {
            std::cerr << "Trace Collector Root Context does not exist"
                      << std::endl;
        }
        catch (const CdmwNamingAndRepository::NoNameDomain&) {
            std::cerr << "Trace Collector Name Domain does not exist \n"
                      << collector_domain_str.c_str()
                      << std::endl;
        }
        catch (const CdmwNamingAndRepository::InvalidName&) {
            std::cerr << "Trace Collector Name Domain has illegal form \n"
                      << collector_domain_str.c_str()
                      << std::endl;
        }
        catch (const CosNaming::NamingContext::InvalidName&) {
            std::cerr << "Trace Collector Name invalid \n"
                      << collector_domain_str.c_str()
                      << std::endl;
        }
        catch (const Cdmw::Common::TypeMismatchException& e) {
            std::cerr << "TypeMismatchException (Not a NamingContext) <"
                      << e.what() << ">" << std::endl;
        }
      } // options.miop_corbaloc == ""

        // create a new POA
        // Trace Collector framework use the following POA policies
        //
        // Lifespan policy            = PERSISTENT
        // Id Assignment policy       = USER_ID
        // Id Uniqueness policy       = UNIQUE_ID
        // Implicit Activation policy = NO_IMPLICIT_ACTIVATION
        // Request Processing policy  = USE_ACTIVE_OBJECT_MAP_ONLY;
        // Servant Retention policy   = RETAIN;
        //
        CORBA::PolicyList policies;
        policies.length(6);

        policies[0] =
            root_POA->create_id_assignment_policy(PortableServer::USER_ID);
        policies[1] =
            root_POA->create_lifespan_policy(PortableServer::PERSISTENT);
        policies[2] =
            root_POA->create_servant_retention_policy(PortableServer::RETAIN);
        policies[3] =
            root_POA->create_id_uniqueness_policy(PortableServer::UNIQUE_ID);
        policies[4] =
            root_POA->create_request_processing_policy(
                PortableServer::USE_ACTIVE_OBJECT_MAP_ONLY);
        policies[5] =
            root_POA->create_implicit_activation_policy(
                PortableServer::NO_IMPLICIT_ACTIVATION);

        Cdmw::OrbSupport::StrategyList poa_strategies;

        // Strategies with MULTITHREAD policy will accept Single Thread and
        // Multi Thread strategies. However, SERIALIZE policy will only allow
        // Single Thread policy.

        PortableServer::POAManager_var poa_manager = root_POA->the_POAManager();
        PortableServer::POA_var collectors_POA = PortableServer::POA::_nil();

        collectors_POA = Cdmw::OrbSupport::OrbSupport::create_POA (
                                                 root_POA,
                                                 options.collector_name.c_str(),
                                                 poa_manager.in(),
                                                 policies,
                                                 poa_strategies);

        // reference to collectors contacted via MIOP
        CdmwTrace::Collector_var miop_collectors_ref = CdmwTrace::Collector::_nil();
        if ( options.miop_corbaloc != "" )
        {
            try 
            {
                // create a group reference for collectors
                CORBA::String_var corbaStrColl =
                  CORBA::string_dup(options.miop_corbaloc.c_str());
                CORBA::Object_var groupColl =
                  orb->string_to_object(corbaStrColl.in());
                miop_collectors_ref =
                  CdmwTrace::Collector::_unchecked_narrow(groupColl.in());
		std::cout << "miop_collectors_ref is set" << std::endl;
            }
            catch (const CORBA::BAD_PARAM&) {
                std::cerr << "Verify your MIOP adress" << std::endl;
                std::cerr <<
                  "Syntax is of kind:" <<
                  " corbaloc:miop:1.0@1.0-aDomain-AAAA/BBB.CCC.DDD.EEE:FFFF"
                  << std::endl;
                std::cerr <<
                  "( Where BBB.CCC.DDD.EEE must be 239.255.0.0 to 239.255.255.255,"
                  << std::endl;
                std::cerr <<
                "  The range 224.255.0.0 to 238.255.255.255 should also be valid )"
                  << std::endl;
                throw;
            }
        }

        // create and initialize the collector manager.
        collector_ref = LocalCollectorMngr::Init (
					     orb,
                                             collectors_POA.in(),
                                             global_collectors,
					     options,
                                             OS::LOCAL_TIME,
                                             miop_collectors_ref.in());
        

        // export the object reference to a file
        CORBA::String_var ref_string = orb->object_to_string(collector_ref.in());
        std::ofstream os("traceCollector.ior");
        os << ref_string.in();
        os.close();


        status = SUCCESS;

        // Register Collector object in Naming & repository
        // under AdminRootContext (using naming interface).
        // trace collector location:
        // <hostname>/SERVICES/TRACE/COLLECTORS/collector_name
        
        try 
        {
            if (status == SUCCESS) 
            {
                status = FAILURE;

                // build the cosnaming of collector
                CosNaming::Name collector_name_;
                collector_name_.length(5);
                collector_name_[0].id =
                    CORBA::string_dup(OS::get_hostname().c_str());
                collector_name_[1].id = CORBA::string_dup("SERVICES");
                collector_name_[2].id = CORBA::string_dup("TRACE");
                collector_name_[3].id = CORBA::string_dup("COLLECTORS");
                collector_name_[4].id = CORBA::string_dup(options.collector_name.c_str());

                // get the admin root context
                CosNaming::NamingContext_var adminRootCtx =
                               p_repository_ref->resolve_root_context(
                                               Locations::ADMIN_ROOT_CONTEXT_ID);

                Cdmw::CommonSvcs::Naming::NamingInterface ni(adminRootCtx.in());
                
                // bind the local collector
                ni.bind(ni.to_string(collector_name_).c_str(),
                        collector_ref.in(),
                        true);

                status = SUCCESS;
            }
        }
        catch (const CdmwNamingAndRepository::Repository::NoRootContext&) {
            std::cerr << "Trace Collector Root Context does not exist"
                      << std::endl;
        }
        catch (const CdmwNamingAndRepository::NoNameDomain&) {
            std::cerr << "Trace Collector Name Domain does not exist"
                      << std::endl;
        }
        catch (const CdmwNamingAndRepository::InvalidName&) {
            std::cerr << "Trace Collector Name Domain has illegal form"
                      << std::endl;
        }
        
        // create the process agent to allow process stopping
        CollectorInit::create_process_agent (false, orb, root_POA);

        // path of trace producer extracted from arguments list
        std::string trace_producer_object_path = "";

        // Register the Collector object to each Trace Producer passed as argument
        try 
        {
            if (status == SUCCESS) 
            {
                // set TraceProducer location under admin root context
                // Pattern is :
                //   "[<host_name>/<application_name>/<process_name>]/TRACE/TraceProducer"

                // Get trace producer root naming context
                CosNaming::NamingContext_var producer_root_nc =
                          p_repository_ref->resolve_root_context(
                                            trace_producer_context.c_str());

                // NamingInterface on Root context
                Cdmw::CommonSvcs::Naming::NamingInterface
                producer_ni(producer_root_nc.in());

                typedef Cdmw::CommonSvcs::Naming::NamingUtil<CdmwTrace::TraceProducer> Util;

                int producer_inx = options.trace_producer_inx;
                
                while (producer_inx < argc) 
                {
                    // Get the path of the Trace Producer object
                    trace_producer_object_path = argv[producer_inx];

                    // Build the complete name of the Trace Producer object
                    std::string producer_context_name = trace_producer_object_path;
                    //trace_producer_object_path contains the complete object name
                    //  producer_context_name += "/TRACE/TraceProducer";

                    try 
                    {
                        // get TraceProducer object using Root context
                        // NamingInterface
                        CdmwTrace::TraceProducer_var traceProducer =
                            Util::resolve_name(producer_ni,
                                               producer_context_name);

                        if (CORBA::is_nil (traceProducer.in())) 
                        {
                            std::cerr << trace_producer_object_path
                                      << " is not a valid trace producer"
                                      << std::endl;
                        }
                        else 
                        {
                            // register the collector in the trace producer
                            traceProducer->register_collector(collector_ref.in(),
                                                              options.collector_name.c_str(),
                                                              options.collector_name.c_str());
                        }
                    }
                    catch (const CosNaming::NamingContext::NotFound&) {
                        std::cerr << "Trace Producer Naming Context NotFound : "
                                  << trace_producer_object_path.c_str()
                                  << std::endl;
                    }
                    catch (const CosNaming::NamingContext::CannotProceed&) {
                        std::cerr
                            << "Unexpected Error (CannotProceed exception)!"
                            << std::endl;
                    }
                    catch (const CosNaming::NamingContext::InvalidName&) {
                        std::cerr << "Invalid Name <"
                                  << trace_producer_object_path << ">!"
                                  << std::endl;
                    }
                    catch (const Cdmw::Common::TypeMismatchException& e) {
                        std::cerr << "TypeMismatchException "
                                  << "(Not a Trace Producer object) <"
                                  << e.what() << ">" << std::endl;
                    }
                    catch (const CORBA::Exception& ex) {
                        std::cerr << "Trace Producer access error : "
                                  << trace_producer_object_path.c_str()
                                  << std::endl << ex << std::endl;
                    }

                    // get next Trace Producer
                    producer_inx++;
                }
            }
        }
        catch (const CdmwNamingAndRepository::Repository::NoRootContext&) {
            std::cerr << "Trace Root Context does not exist" << std::endl;
            status = FAILURE;
        }


    }
    catch (const std::bad_alloc&) {
        std::cerr << "Process Initialisation : Bad Allocation Exception \n" ;
    }
    catch (const Cdmw::Exception& ex) {
        std::cerr << ex.what() << std::endl;
    }
    catch (const CORBA::SystemException &ex) {
        std::cerr << "CORBA::SystemException" << ex << std::endl;
    }
    catch(const CORBA::Exception& ex) {
        std::cerr << "CORBA::Exception:" << ex << std::endl;
    }

    return status;
}


//
// end of processing
//
void end_local_collector(void)
{
    CollectorMngr::Cleanup();
}

void exceptionHandler(void)
{
    std::cout << "Local Trace Collector : UNEXPECTED EXCEPTION HANDLER" << std::endl;
}




void exit_handler(int sig) 
{
    if (!CORBA::is_nil(orb.in())) 
    {
        try 
        {
            orb->shutdown(false);
        } catch (const CORBA::SystemException& e) {
            std::cerr << "Error while shuting ORB down in exit_handler:\n"
                      << e << " - minor code: " << e.minor() << std::endl;
        }
    }
}


//
// main module
//
int main(int argc, char* argv[])
{
    // for international char management
    // set for all locale categories the traditional unix system behavior
    ::setlocale (LC_ALL,"C");
    
#   ifndef _MSC_VER    

    std::set_unexpected(&exceptionHandler);

    // VC++ contrary to the clause 18.6 of the C++ standard, 
     //  set_unexpected, is not inside the std:: namespace
#   else

    set_unexpected(&exceptionHandler);

#   endif  



    using namespace Cdmw::OsSupport;
    
    // get the collector options
    CollectorInit::CollectorOptions options;
    bool arg_rebuilt = false;
    char** new_argv = argv;
    int new_argc = argc;
    int arg_ini_inx = argc;

    int result = CollectorInit::get_collector_options (false, argc, argv, options,
                                                       new_argc, new_argv,
                                                       DFLT_SERVICE_PORT,
                                                       DFLT_PROCESS_PORT);
    if (result == FAILURE)
    {
        return FAILURE;
    }
    else if (result == EXIT)
    {
        return SUCCESS;
    }
    
    // if new arguments has been added                                                   
    if (new_argc > argc)
    {
        arg_rebuilt = true;
    }
    
  
    std::cout << "Local Trace Collector " << options.collector_name.c_str() 
              << " is listening on port " << options.service_port << "\n"; 
    std::cout << "   (" << Cdmw::Common::Options::PROCESS_PORT_OPTION
              << " is set to " << options.process_port << ")" << "\n";
    std::cout << std::endl;


    int status = SUCCESS;
    
    try 
    {
        PortableGroup::GOA_var root_GOA;        
        CdmwNamingAndRepository::Repository_var repository_ref =
                CdmwNamingAndRepository::Repository::_nil();
                        
        bool use_miop=true;
        if (options.miop_corbaloc == "") { use_miop = false; }

        // init the orb and platforminterface
        // (exception is raised in case of error)
        CollectorInit::init_processing (
                      new_argc, new_argv,
                      options.platform_managed,
                      options.naming_repos_url,
                      orb.out(), root_GOA.out(),
                      repository_ref.out(),
                      use_miop);
        
        // set Trace producer context (under admin root context)
        std::string trace_producer_context = 
                Cdmw::Common::Locations::ADMIN_ROOT_CONTEXT_ID;              

        // Init the collector
        status = init_local_collector (orb.in(), 
                                       root_GOA.in(),
                                       options,
                                       repository_ref.in(),
                                       trace_producer_context, 
                                       argc, argv);

        if (status == SUCCESS) 
        {
            struct sigaction action;
            action.sa_handler = exit_handler;
            sigemptyset(&action.sa_mask);
            sigaction(SIGTERM, &action, NULL);
            sigaction(SIGINT, &action, NULL);
            sigaction(SIGQUIT, &action, NULL);
            sigaction(SIGABRT, &action, NULL);

            // run processing
            status = CollectorInit::run_processing (orb.in(), options.interactive_mode);
        }

        // end of process
        end_local_collector();

        if (options.platform_managed) 
        {
            // cleanup the platform interface
            PlatformInterface::Cleanup();
        }
    }
    catch (const Cdmw::Exception& ex) {
        std::cerr << ex.what() << std::endl;
        status = FAILURE;
    }
    catch(const CORBA::Exception& ex) {
        std::cerr << ex << std::endl;
        status = FAILURE;
    }
    catch (...) {
        std::cerr << "unknown exception in main" << std::endl;
        status = FAILURE;
    }

    // end of corba processing

    if(!CORBA::is_nil(orb.in())) 
    {
        try 
        {
            Cdmw::OrbSupport::OrbSupport::ORB_cleanup(orb.in());
            orb->destroy();
        }
        catch(const CORBA::Exception& ex) 
        {
            std::cerr << ex << std::endl;
            status = FAILURE;
        }
    }

    // if program arguments table has been rebuilt
    if (arg_rebuilt) 
    {
        // release any allocated memory
        for (int i=arg_ini_inx ; i < new_argc; ++i) 
        {
            delete[] new_argv[i];
        }

        delete[] new_argv;
    }

    return status;
}

